# Standard imports
import functools
import math
import csv
import io
import time
import itertools as it
from collections import defaultdict
import copy
import sys
import string
import urllib.request  # STRANGE: CANNOT IMPORT URLLIB ALONE
from logging import DEBUG

# dependency
import cachetools

# Qt imports
from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *

# Custom imports
from cutevariant.core.querybuilder import build_full_sql_query, fields_to_vql
from cutevariant.core import sql

from cutevariant.core import command as cmd
from cutevariant.gui import plugin, FIcon, formatter, style
from cutevariant.gui.sql_thread import SqlThread
from cutevariant.gui.widgets import MarkdownEditor
import cutevariant.commons as cm

LOGGER = cm.logger()


class VariantModel(QAbstractTableModel):
    """VariantModel is a Qt model class which contains variant data from SQL DB.

    It loads paginated data and create an interface for Qt views and controllers.
    The model can group variants by (chr,pos,ref,alt).

    See Qt model/view programming for more information
    https://doc.qt.io/qt-5/model-view-programming.html

    Variants are stored internally as a list of variants.
    By default, there is only one variant per row until a user selects a field
    from annotations or from multiple samples.
    Duplicated variants will be displayed in this case. It is advised to use the
    button "group" in the GUI to easily split between specific fields and common
    fields.

    Signals:
        variant_loaded(bool): Emit when variant are loaded 
        count_loaded(bool): Emit when total count are loaded 
        error_raised(str): Emit message when threads or something else encounter errors
    """

    # emit when variant results is loaded
    variant_loaded = Signal()
    variant_is_loading = Signal(bool)

    # emit when toutal count is loaded
    count_loaded = Signal()
    count_is_loading = Signal(bool)

    # Emit when all load has started
    load_started = Signal()

    # Emit when all data ( count + variant) has finished
    load_finished = Signal()

    error_raised = Signal(str)
    interrupted = Signal()

    DEFAUT_CACHE_SIZE = 1_048_576 * 32  # Default cache size of 32 Mo

    def __init__(self, conn=None, parent=None):
        super().__init__()
        self.limit = 50
        self.page = 1  #
        self.total = 0
        self.variants = []
        self.headers = []

        # Cache all database fields and their descriptions for tooltips
        # Field names as keys, descriptions as values
        self.fields_descriptions = None

        self.fields = ["chr", "pos", "ref", "alt"]
        self.filters = dict()
        self.source = "variants"
        self.group_by = []
        self.having = {}
        self.order_by = None
        self.order_desc = False
        self.formatter = None
        self.debug_sql = None
        # Keep after all initialization
        self.conn = conn

        # Thread (1 for getting variant, 1 for getting count variant )
        self._load_variant_thread = SqlThread(self.conn)
        self._load_count_thread = SqlThread(self.conn)

        self._load_variant_thread.started.connect(
            lambda: self.variant_is_loading.emit(True)
        )
        self._load_variant_thread.finished.connect(
            lambda: self.variant_is_loading.emit(False)
        )
        self._load_variant_thread.result_ready.connect(self._on_variant_loaded)
        self._load_variant_thread.error.connect(self.error_raised)

        self._load_count_thread.started.connect(
            lambda: self.count_is_loading.emit(True)
        )
        self._load_count_thread.finished.connect(
            lambda: self.count_is_loading.emit(False)
        )
        self._load_count_thread.result_ready.connect(self._on_count_loaded)

        self._finished_thread_count = 0
        self._user_has_interrupt = False

        # Create results cache because Thread doesn't use the memoization cache from command.py.
        # This is because Thread create a new connection and change the function signature used by the cache.

        self._load_variant_cache = cachetools.LFUCache(
            maxsize=self.DEFAUT_CACHE_SIZE, getsizeof=sys.getsizeof
        )
        self._load_count_cache = cachetools.LFUCache(maxsize=1000)

    @property
    def conn(self):
        """ Return sqlite connection """
        return self._conn

    @conn.setter
    def conn(self, conn):
        """ Set sqlite connection """
        self._conn = conn
        if conn:
            # Note: model is initialized with None connection during start
            # Cache DB fields descriptions
            self.fields_descriptions = {
                field["name"]: field["description"]
                for field in sql.get_fields(self.conn)
            }
            LOGGER.debug("Init async thread")

            # Clear cache with new connection
            self.clear_all_cache()

            # Init Runnables (1 for each query type)
            self._load_variant_thread.conn = conn
            self._load_count_thread.conn = conn

            # self._load_count_thread.error.connect(self._on_error)

    def rowCount(self, parent=QModelIndex()):
        """Overrided : Return children count of index"""
        # If parent is root

        return len(self.variants)

    def columnCount(self, parent=QModelIndex()):
        """Overrided: Return column count of parent .

        Parent is not used here 
        """

        #  Check integrity for unit test
        if parent == QModelIndex():
            return len(self.headers)
        return 0

    def clear_all_cache(self):
        """ clear cache """
        self.clear_count_cache()
        self.clear_variant_cache()

    def clear_variant_cache(self):
        self._load_variant_cache.clear()

    def clear_count_cache(self):
        self._load_count_cache.clear()

    def cache_size(self):
        """ Return total cache size """
        return self._load_variant_cache.currsize

    def max_cache_size(self):
        return self._load_variant_cache.maxsize

    def is_variant_loading(self):
        return self._load_variant_cache.isRunning()

    def is_count_loading(self):
        return self._load_count_thread.isRunning()

    def clear(self):
        """Reset the current model

            - clear variants list
            - total of variants is set to 0
            - emit variant_loaded signal
        """
        self.beginResetModel()
        self.variants.clear()
        self.total = 0
        self.endResetModel()
        self.variant_loaded.emit()

    def data(self, index: QModelIndex(), role=Qt.DisplayRole):
        """Overrided: return index data according role.
        This method is called by the Qt view to get data to display according Qt role.

        Params:
            index (QModelIndex): index from where your want to get data
            role (Qt.ItemDataRole): https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum

        Examples:
            index = model.index(row=10, column = 1)
            chromosome_value = model.data(index)
        """

        # Avoid error
        if not index.isValid():
            return

        if self.variants and self.headers:

            column_name = self.headers[index.column()]

            # ---- Display Role ----
            if role == Qt.DisplayRole:
                value = self.variant(index.row())[column_name]
                if value is None:
                    return "NULL"
                else:
                    return str(self.variant(index.row())[column_name])

    def headerData(self, section, orientation=Qt.Horizontal, role=Qt.DisplayRole):
        """Overrided: Return column name and display tooltips on headers

        This method is called by the Qt view to display vertical or horizontal header data.

        Params:
            section (int): row or column number depending on orientation
            orientation (Qt.Orientation): Qt.Vertical or Qt.Horizontal
            role (Qt.ItemDataRole): https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum

        Examples:
            # return 4th column name
            column_name = model.headerData(4, Qt.Horizontal)

        Warnings:
            Return nothing if orientation is != from Qt.Horizontal,
            and role != Qt.DisplayRole
        """
        # Display columns headers
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                return self.headers[section]

            if role == Qt.ToolTipRole:
                # Field descriptions on headers
                # Note: fields are set in load()
                if self.fields_descriptions and section != 0:
                    field_name = self.fields[section - 1]
                    return self.fields_descriptions.get(field_name)

    def update_variant(self, row: int, variant: dict):
        """Update a variant at the given row with given content

        Update the variant in the GUI AND in the DB.

        Args:
            row (int): Row id of the variant that will be modified
            variant (dict): Dict of fields to be updated
        """
        # Update in database
        variant_id = self.variants[row]["id"]

        print("variant", variant)

        # Update all variant with same variant_id
        # Use case : When several transcript are displayed
        for row in self.find_row_id_from_variant_id(variant_id):
            print(row)
            left = self.index(row, 0)
            right = self.index(row, self.columnCount() - 1)

            if left.isValid() and right.isValid():
                # Get database id of the variant to allow its update operation
                variant["id"] = self.variants[row]["id"]
                sql.update_variant(self.conn, variant)
                self.variants[row].update(variant)
                self.dataChanged.emit(left, right)

    def find_row_id_from_variant_id(self, variant_id: int) -> list:
        """Find the ids of all rows with the same given variant_id

        Args:
            variant_id (int): Variant sql record id
        Returns:
            (list[int]): ids of rows
        """
        return [
            row_id
            for row_id, variant in enumerate(self.variants)
            if variant["id"] == variant_id
        ]

    def interrupt(self):
        """Interrupt current query if active

        This is a blocking function... 

        call interrupt and wait for the error_raised signals ...
        If nothing happen after 1000 ms, by pass and continue
        If I don't use the dead time, it is waiting for an infinite time
        at startup ... Because at startup, loading is called 2 times.
        One time by the register_plugin and a second time by the plugin.show_event
        """

        interrupted = False

        if self._load_count_thread:
            if self._load_count_thread.isRunning():
                self._user_has_interrupt = True
                self._load_count_thread.interrupt()
                self._load_count_thread.wait(1000)
                interrupted = True

        if self._load_variant_thread:
            if self._load_variant_thread.isRunning():
                self._user_has_interrupt = True
                self._load_variant_thread.interrupt()
                self._load_variant_thread.wait(1000)
                interrupted = True

        if interrupted:
            self.interrupted.emit()

        # # Wait for exception ...
        # if loop:
        #     self.error_raised.connect(loop.quit)
        #     loop.exec_()

    def is_running(self):
        if self._load_variant_thread and self._load_count_thread:
            return (
                self._load_variant_thread.isRunning()
                or self._load_count_thread.isRunning()
            )

        return False

    def load(self):
        """Start async queries to get variants and variant count

        Called by:
            - on_change_query() from the view.
            - sort() and setPage() by the model.

        See Also:
            :meth:`loaded`
        """
        if self.conn is None:
            return

        if self.is_running():
            LOGGER.debug(
                "Cannot load data. Thread is not finished. You can call interrupt() "
            )

        LOGGER.debug("Start loading")

        offset = (self.page - 1) * self.limit

        # Add fields from group by
        # self.clear()  # Assume variant = []
        self.total = 0
        self._finished_thread_count = 0
        # LOGGER.debug("Page queried: %s", self.page)

        print("ORDER", self.order_by, self.order_desc)

        # Store SQL query for debugging purpose
        self.debug_sql = build_full_sql_query(
            self.conn,
            fields=self.fields,
            source=self.source,
            filters=self.filters,
            limit=self.limit,
            offset=offset,
            order_desc=self.order_desc,
            order_by=self.order_by,
            group_by=self.group_by,
            having=self.having,
        )

        # Create load_func to run asynchronously: load variants
        load_func = functools.partial(
            cmd.select_cmd,
            fields=self.fields,
            source=self.source,
            filters=self.filters,
            limit=self.limit,
            offset=offset,
            order_desc=self.order_desc,
            order_by=self.order_by,
            group_by=self.group_by,
            having=self.having,
        )

        # Create count_func to run asynchronously: count variants
        count_function = functools.partial(
            cmd.count_cmd,
            fields=self.fields,
            source=self.source,
            filters=self.filters,
            group_by=self.group_by,
        )

        # Start the run
        self._start_timer = time.perf_counter()

        # Create function HASH for CACHE
        self._count_hash = hash(
            count_function.func.__name__ + str(count_function.keywords)
        )
        self._variant_hash = hash(load_func.func.__name__ + str(load_func.keywords))

        self.load_started.emit()

        # Launch the first thread "count" or by pass it using the cache
        if self._count_hash in self._load_count_cache:
            self._load_count_thread.results = self._load_count_cache[self._count_hash]
            self._on_count_loaded()
        else:
            self._load_count_thread.start_function(count_function)

        # Launch the second thread "count" or by pass it using the cache
        if self._variant_hash in self._load_variant_cache:
            self._load_variant_thread.results = self._load_variant_cache[
                self._variant_hash
            ]
            self._on_variant_loaded()

        else:
            self._load_variant_thread.start_function(lambda conn: list(load_func(conn)))

    def _on_variant_loaded(self):
        """
        Triggered when variant_thread is finished 

        """

        #  Compute time elapsed since loading
        self._end_timer = time.perf_counter()
        self.elapsed_time = self._end_timer - self._start_timer

        self.beginResetModel()
        self.variants.clear()

        # Add fields from group by
        for g in self.group_by:
            if g not in self.fields:
                self.fields.append(g)

        # Save cache
        self._load_variant_cache[
            self._variant_hash
        ] = self._load_variant_thread.results.copy()

        # Load variants
        self.variants = self._load_variant_thread.results
        if self.variants:
            # Set headers of the view
            self.headers = list(self.variants[0].keys())

        # self.total = self._load_count_thread.results["count"]

        self.endResetModel()
        self.variant_loaded.emit()

        #  Test if both thread are finished
        self._finished_thread_count += 1
        if self._finished_thread_count == 2:
            self.load_finished.emit()

    def _on_count_loaded(self):
        """
        Triggered when count_threaed is finished 
        """

        # Save cache
        self._load_count_cache[
            self._count_hash
        ] = self._load_count_thread.results.copy()

        self.total = self._load_count_thread.results["count"]
        self.count_loaded.emit()

        #  Test if both thread are finished
        self._finished_thread_count += 1
        if self._finished_thread_count == 2:
            self.load_finished.emit()

    def hasPage(self, page: int) -> bool:
        """ Return True if <page> exists otherwise return False """
        return (page - 1) >= 0 and (page - 1) * self.limit < self.total

    def setPage(self, page: int):
        """ set the page of the model """
        if self.hasPage(page):
            self.page = page

    def nextPage(self):
        """ Set model to the next page """
        if self.hasPage(self.page + 1):
            self.setPage(self.page + 1)

    def previousPage(self):
        """ Set model to the previous page """
        if self.hasPage(self.page - 1):
            self.setPage(self.page - 1)

    def firstPage(self):
        """ Set model to the first page """
        self.setPage(1)

    def lastPage(self):
        """ Set model to the last page """

        self.setPage(self.pageCount())

    def pageCount(self):
        """ Return total page count """
        return math.ceil(self.total / self.limit)

    def sort(self, column: int, order):
        """Overrided: Sort data by specified column

        column (int): column id
        order (Qt.SortOrder): Qt.AscendingOrder or Qt.DescendingOrder

        """
        if column < self.columnCount():
            field = self.fields[column - 1]
            self.order_by = field
            self.order_desc = order == Qt.DescendingOrder
            self.load()

    def variant(self, row: int) -> dict:
        """Return variant data according index"""
        return self.variants[row]

    def is_variant_loading(self):
        if self._load_variant_thread:
            return self._load_variant_thread.isRunning()
        else:
            return False

    def is_count_loading(self):
        if self._load_count_thread:
            return self._load_count_thread.isRunning()
        else:
            return False


class VariantDelegate(QStyledItemDelegate):
    """Specify the aesthetic (style and color) of variants displayed on a view"""

    def __init__(self, parent=None):
        super().__init__(parent)

        self.formatter = None

    def paint(self, painter, option, index):
        """Paint with formatter if defined"""

        if self.formatter is None:
            return super().paint(painter, option, index)

        # Draw selections
        if option.state & QStyle.State_Enabled:
            bg = (
                QPalette.Normal
                if option.state & QStyle.State_Active
                or option.state & QStyle.State_Selected
                else QPalette.Inactive
            )
        else:
            bg = QPalette.Disabled

        if option.state & QStyle.State_Selected:
            painter.fillRect(option.rect, option.palette.color(bg, QPalette.Highlight))

        # Draw formatters
        option.rect = option.rect.adjusted(
            3, 0, 0, 0
        )  # Don't know why I need to adjust the left margin .. .
        self.formatter.paint(painter, option, index)


class LoadingTableView(QTableView):
    """Movie animation displayed on VariantView for long SQL queries executed
    in background.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self._is_loading = False

    def paintEvent(self, event: QPainter):

        if self.is_loading():
            painter = QPainter(self.viewport())

            painter.drawText(
                self.viewport().rect(), Qt.AlignCenter, self.tr("Loading ...")
            )

        else:
            super().paintEvent(event)

    def start_loading(self):
        self._is_loading = True
        self.viewport().update()

    def stop_loading(self):
        self._is_loading = False
        self.viewport().update()

    def is_loading(self):
        return self._is_loading


class VariantView(QWidget):
    """A Variant view with controller like pagination

    Attributes:

        row_to_be_selected (int/None): (optional) Left groupby pane only:
            At the end of the :meth:`loaded` method, the first line is selected
            in order to refresh the current variant in the other pane.
            TL,DR: Select the first row if in grouped mode => refresh the right pane.

    Signals:
        error_raised (str): Emit message when async runnables encounter errors
        no_variant: Emitted when there is no variant to display.
            Used by left pane to clear right pane.
    """

    error_raised = Signal(str)
    no_variant = Signal()

    def __init__(self, parent=None, show_popup_menu=True):
        """
        Args:
            parent: parent widget
            show_popup_menu (boolean, optional: If False, disable the context menu
                over variants. For example the group pane should be disabled
                in order to avoid partial/false informations to be displayed
                in this menu.
                Hacky Note: also used to rename variants to groups in the page box...
        """
        super().__init__(parent)

        self.parent = parent
        self.show_popup_menu = show_popup_menu
        self.view = LoadingTableView()
        self.bottom_bar = QToolBar()

        # self.view.setFrameStyle(QFrame.NoFrame)
        self.view.setAlternatingRowColors(True)
        self.view.horizontalHeader().setStretchLastSection(True)
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.view.verticalHeader().hide()

        self.view.setSortingEnabled(True)
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.ExtendedSelection)
        ## self.view.setIndentation(0)
        self.view.setIconSize(QSize(22, 22))
        self.view.horizontalHeader().setSectionsMovable(True)

        # Setup model
        self.model = VariantModel()
        self.view.setModel(self.model)

        # Setup delegate
        self.delegate = VariantDelegate()
        self.view.setItemDelegate(self.delegate)

        # setup toolbar
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        self.page_box = QLineEdit()
        self.page_box.setValidator(QIntValidator())
        self.page_box.setFixedWidth(50)
        self.page_box.setValidator(QIntValidator())

        # Display nb of variants/groups and pages
        self.info_label = QLabel()
        self.time_label = QLabel()
        self.cache_label = QLabel()
        self.loading_label = QLabel()
        self.loading_label.setMovie(QMovie(cm.DIR_ICONS + "loading.gif"))

        self.bottom_bar.addAction(FIcon(0xF0A30), "sql", self.on_show_sql)
        self.bottom_bar.addWidget(self.time_label)
        self.bottom_bar.addWidget(self.cache_label)
        self.bottom_bar.addSeparator()
        self.bottom_bar.addWidget(spacer)

        # Add loading action and store action
        self.bottom_bar.addWidget(self.info_label)
        self.loading_action = self.bottom_bar.addWidget(self.loading_label)

        self.bottom_bar.setIconSize(QSize(16, 16))
        self.bottom_bar.setMaximumHeight(30)
        self.bottom_bar.setContentsMargins(0, 0, 0, 0)

        self.pagging_actions = []

        action = self.bottom_bar.addAction(FIcon(0xF0600), "<<", self.on_page_clicked)
        action.setShortcut(Qt.CTRL + Qt.Key_Left)
        action.setAutoRepeat(False)
        action.setToolTip(self.tr("First page (%s)" % action.shortcut().toString()))

        self.pagging_actions.append(action)

        action = self.bottom_bar.addAction(FIcon(0xF0141), "<", self.on_page_clicked)
        action.setShortcut(QKeySequence(Qt.Key_Left))
        action.setAutoRepeat(False)
        action.setToolTip(self.tr("Previous page (%s)" % action.shortcut().toString()))
        self.pagging_actions.append(action)

        self.bottom_bar.addWidget(self.page_box)
        self.page_box.returnPressed.connect(self.on_page_changed)

        action = self.bottom_bar.addAction(FIcon(0xF0142), ">", self.on_page_clicked)
        action.setShortcut(QKeySequence(Qt.Key_Right))
        action.setAutoRepeat(False)
        action.setToolTip(self.tr("Next page (%s)" % action.shortcut().toString()))
        self.pagging_actions.append(action)

        action = self.bottom_bar.addAction(FIcon(0xF0601), ">>", self.on_page_clicked)
        action.setShortcut(Qt.CTRL + Qt.Key_Right)
        action.setAutoRepeat(False)
        action.setToolTip(self.tr("Last page (%s)" % action.shortcut().toString()))
        self.pagging_actions.append(action)

        self.fav_action = QAction(FIcon(0xF00C0), self.tr("Toggle favorite "))
        self.fav_action.triggered.connect(lambda: self.update_favorites())
        self.fav_action.setShortcut(QKeySequence(Qt.Key_Space))
        self.fav_action.setShortcutContext(Qt.WidgetShortcut)
        self.fav_action.setAutoRepeat(False)
        self.fav_action.setToolTip(
            self.tr("Toggle favorite (%s)" % self.fav_action.shortcut().toString())
        )
        self.view.addAction(self.fav_action)

        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.addWidget(self.view)
        main_layout.addWidget(self.bottom_bar)
        self.setLayout(main_layout)

        # Async stuff
        # broadcast focus signal
        self.row_to_be_selected = None
        # Get the status of async load: started/finished
        # self.model.loading.connect(self._set_loading)
        # Queries are finished (yes its redundant with loading signal...)
        self.model.variant_loaded.connect(self._on_variant_loaded)
        self.model.count_loaded.connect(self._on_count_loaded)

        self.model.count_is_loading.connect(self.set_tool_loading)
        self.model.variant_is_loading.connect(self.set_view_loading)

        # Connect errors from async runnables
        self.model.error_raised.connect(self.error_raised)
        #  connect double clicke
        self.view.doubleClicked.connect(self.on_double_clicked)

    # def _set_loading(self, active=True):
    #     """Slot to obtain the status of async load: started/finished

    #     Start/Stop the movie animation on the view.

    #     Keyword Args:
    #         active (bool): True if loaded, False if finished
    #     """
    #     self.setDisabled(active)
    #     if active:
    #         QTimer.singleShot(1000, self.start_loading_after_delay)
    #     else:
    #         self.view.stop_loading()

    # def start_loading_after_delay(self):
    #     self.setDisabled(self.model.is_loading)

    #     if self.model.is_loading:
    #         self.view.start_loading()
    #     else:
    #         self.view.stop_loading()

    def setModel(self, model: VariantModel):
        self.model = model
        self.view.setModel(model)

    def load(self, reset_page=False):

        if reset_page:
            self.model.page = 1
            self.model.order_by = None

        self.model.interrupt()

        self.model.load()

    def _on_variant_loaded(self):
        """Slot called when async queries from the model are finished
        (especially count of variants for page box).

        Signals:
            Emits no_variant signal.
        """
        # if self.row_to_be_selected is not None:
        #     # Left groupby pane only:
        #     # Select by default the first line in order to refresh the
        #     # current variant in the other pane
        #     if self.model.rowCount():
        #         self.select_row(0)
        #     else:
        #         self.no_variant.emit()

        self.time_label.setText(str(" Executed in %.2gs " % (self.model.elapsed_time)))
        cache = cm.bytes_to_readable(self.model.cache_size())
        max_cache = cm.bytes_to_readable(self.model.max_cache_size())
        self.cache_label.setText(str(" Cache {} of {}".format(cache, max_cache)))
        if LOGGER.getEffectiveLevel() != DEBUG:
            self.view.setColumnHidden(0, True)
        self.view.scrollToTop()

        #  Select first row
        if self.model.rowCount():
            self.select_row(0)

    def _on_count_loaded(self):

        self.page_box.clear()
        if self.model.pageCount() - 1 == 0:
            self.set_pagging_enabled(False)
        else:
            # self.page_box.addItems([str(i) for i in range(self.model.pageCount())])
            self.page_box.validator().setRange(1, self.model.pageCount())
            self.page_box.setText(str(self.model.page))
            self.set_pagging_enabled(True)

        if not self.show_popup_menu:
            # Yes it's hacky... but left pane doesn't show variants
            text = self.tr("{} group(s) {} page(s)")
        else:
            text = self.tr("{} line(s) {} page(s)")

        print("loaded")
        self.info_label.setText(text.format(self.model.total, self.model.pageCount()))

    def set_formatter(self, formatter):
        self.delegate.formatter = formatter
        self.view.reset()

    @property
    def conn(self):
        return self.model.conn

    @conn.setter
    def conn(self, _conn):
        self.model.conn = _conn

    @property
    def fields(self):
        return self.model.fields

    @fields.setter
    def fields(self, _fields):
        self.model.fields = _fields

    @property
    def source(self):
        return self.model.source

    @source.setter
    def source(self, _source):
        self.model.source = _source

    @property
    def filters(self):
        return self.model.filters

    @filters.setter
    def filters(self, _filters):
        self.model.filters = _filters

    @property
    def group_by(self):
        return self.model.group_by

    @group_by.setter
    def group_by(self, _group_by):
        self.model.group_by = _group_by

    def on_page_clicked(self):

        action_text = self.sender().text()

        if action_text == "<<":
            fct = self.model.firstPage

        if action_text == ">>":
            fct = self.model.lastPage

        if action_text == "<":
            fct = self.model.previousPage

        if action_text == ">":
            fct = self.model.nextPage

        fct()
        self.load()

    def on_page_changed(self):
        """Slot called when page_box is modified and the user has pressed return key

        The validator is ok if this slot is called.
        """
        if self.page_box.text():
            page = int(self.page_box.text())
            self.model.setPage(page)
            self.load()

    def on_show_sql(self):
        """Display debug sql query"""
        msg_box = QMessageBox()
        msg_box.setWindowTitle("SQL debug")
        msg_box.setText(self.model.debug_sql)
        explain_query = []
        for rec in self.conn.execute("EXPLAIN QUERY PLAN " + self.model.debug_sql):
            explain_query.append(str(dict(rec)))

        msg_box.setDetailedText("\n".join(explain_query))

        msg_box.exec_()

    def set_pagging_enabled(self, active=True):
        self.page_box.setEnabled(active)
        for action in self.pagging_actions:
            action.setEnabled(active)

    def set_view_loading(self, active=True):
        self.view.setDisabled(active)

        def show_loading_if_loading():
            if self.model.is_variant_loading():
                self.view.start_loading()

        if active:
            QTimer.singleShot(2000, show_loading_if_loading)
        else:
            self.view.stop_loading()

    def set_tool_loading(self, active=True):

        if active:
            self.info_label.setText(
                self.tr("Counting all variants. This can take a while ... ")
            )
            self.loading_action.setVisible(True)
            self.loading_label.movie().start()
        else:
            self.loading_label.movie().stop()
            self.loading_action.setVisible(False)
            # self.info_label.setText("")

        self.bottom_bar.setDisabled(active)

    def set_loading(self, active=True):

        self.set_view_loading(active)
        self.set_tool_loading(active)

    def _get_links(self) -> list:
        """Get links from settings

        Return list of links from QSettings 

        Exemples:
            {
            "name":"google",
            "url": "http://www.google.fr/q={}",
            "is_browser": True   # Open with browser  
            "is_default": True   # is a default action 
            }

        """
        links = []
        settings = QSettings()
        settings.beginGroup("variant_view")
        size = settings.beginReadArray("links")
        for index in range(size):
            settings.setArrayIndex(index)

            links.append(
                {
                    "name": settings.value("name"),
                    "url": settings.value("url"),
                    "is_default": settings.value("is_default", 0, type=bool),
                    "is_browser": settings.value("is_browser", 0, type=bool),
                }
            )

        settings.endArray()
        settings.endGroup()

        return links

    def contextMenuEvent(self, event: QContextMenuEvent):
        """Override: Show contextual menu over the current variant"""
        if not self.show_popup_menu:
            return

        menu = QMenu(self)
        pos = self.view.viewport().mapFromGlobal(event.globalPos())
        current_index = self.view.indexAt(pos)

        if not current_index.isValid():
            return

        current_variant = self.model.variant(current_index.row())
        full_variant = sql.get_one_variant(self.conn, current_variant["id"])
        # Update variant with currently displayed fields (not in variants table)
        full_variant.update(current_variant)

        # Copy action: Copy the variant reference ID in to the clipboard
        formatted_variant = "{chr}:{pos}-{ref}-{alt}".format(**full_variant)
        menu.addAction(
            FIcon(0xF014C),
            formatted_variant,
            functools.partial(
                QApplication.instance().clipboard().setText, formatted_variant
            ),
        )

        # Create favorite action
        fav_action = menu.addAction(
            self.tr("&Unmark favorite")
            if bool(full_variant["favorite"])
            else self.tr("&Mark as favorite")
        )
        fav_action.setCheckable(True)
        fav_action.setChecked(bool(full_variant["favorite"]))
        fav_action.toggled.connect(self.update_favorites)

        # Create classication action
        class_menu = menu.addMenu(self.tr("Classification"))
        for key, value in cm.CLASSIFICATION.items():

            action = class_menu.addAction(FIcon(cm.CLASSIFICATION_ICONS[key]), value)
            action.setData(key)
            on_click = functools.partial(self.update_classification, current_index, key)
            action.triggered.connect(on_click)

        # Create external links
        links_menu = menu.addMenu(self.tr("External links"))

        # Display only exec_ternal links with placeholders that can be mapped
        for link in self._get_links():
            url = self._create_url(link["url"], full_variant)
            if url:
                func_slot = functools.partial(self._open_url, url, link["is_browser"])
                action = links_menu.addAction(link["url"], func_slot)

        # Comment action
        on_edit = functools.partial(self.edit_comment, current_index)
        menu.addAction(self.tr("&Edit comment ..."), on_edit)

        # Edit menu
        menu.addSeparator()
        menu.addAction(
            FIcon(0xF018F), self.tr("&Copy"), self.copy_to_clipboard, QKeySequence.Copy
        )
        menu.addAction(
            FIcon(0xF0486),
            self.tr("&Select all"),
            self.select_all,
            QKeySequence.SelectAll,
        )

        # Display
        menu.exec_(event.globalPos())

    def _open_url(self, url: QUrl, in_browser=False):

        if in_browser:
            QDesktopServices.openUrl(url)

        else:
            urllib.request.urlopen(url.toString(), timeout=10)

    def _create_url(self, format_string: str, variant: dict) -> QUrl:
        """Create a link from a format string and a variant data 
        
        Args:
            format_string (str): a string with format group like : http://www.google.fr?q={chr}
            variant (dict): a variant dict returned by sql.get_one_variant.
        
        Returns:
            QUrl: return url or return None
        
        """
        field_names = {
            name
            for _, name, _, _ in string.Formatter().parse(format_string)
            if name is not None
        }
        if field_names & variant.keys():
            # Full or partial mapping => accepted link
            return QUrl(format_string.format(**variant), QUrl.TolerantMode)
        else:
            return format_string

    def update_favorites(self, checked: bool = None):
        """Update favorite status of multiple selected variants

        if checked is None, toggle favorite

        Warnings:
            BE CAREFUL with this code, we try to limit useless SQL queries as
            much as possible.
        """

        # Do not update the same variant multiple times
        unique_ids = set()

        for index in self.view.selectionModel().selectedRows():
            if not index.isValid():
                continue

            # Get variant id
            variant = self.model.variants[index.row()]
            variant_id = variant["id"]

            if variant_id in unique_ids:
                continue
            unique_ids.add(variant_id)

            # Update GUI + DB
            if checked is None:
                #  Toggle checked box
                update_data = {"favorite": int(not bool(variant["favorite"]))}
            else:
                update_data = {"favorite": int(checked)}

            self.model.update_variant(index.row(), update_data)

    def update_classification(self, index: QModelIndex, value=3):
        """Update classification level of the variant at the given index"""
        if index.isValid():
            update_data = {"classification": int(value)}
            self.model.update_variant(index.row(), update_data)

    def edit_comment(self, index: QModelIndex):
        """Allow a user to add a comment for the selected variant"""
        if not index.isValid():
            return

        # Get comment from DB
        variant_data = sql.get_one_variant(
            self.model.conn, self.model.variant(index.row())["id"]
        )
        comment = variant_data["comment"] if variant_data["comment"] else ""

        editor = MarkdownEditor(default_text=comment)
        if editor.exec_() == QDialog.Accepted:
            # Save in DB
            self.model.update_variant(index.row(), {"comment": editor.toPlainText()})

            # Request a refresh of the variant_info plugin
            self.parent.mainwindow.refresh_plugin("variant_info")

    def select_all(self):
        """Select all variants in the view"""
        self.view.selectAll()

    def select_row(self, row):
        """Select the row with the given index

        Called for left pane by :meth:`loaded`.
        """
        index = self.view.model().index(row, 0)
        self.view.selectionModel().setCurrentIndex(
            index, QItemSelectionModel.SelectCurrent | QItemSelectionModel.Rows
        )

    def copy_to_clipboard(self):
        """Copy the selected variant(s) into the clipboard

        The output data is formated in CSV (delimiter is `\t`)
        """
        # In memory file
        output = io.StringIO()
        # Use CSV to securely format the data
        writer = csv.DictWriter(output, delimiter="\t", fieldnames=self.model.headers)
        writer.writeheader()
        for index in self.view.selectionModel().selectedRows():
            # id col is not wanted
            variant = dict(self.model.variant(index.row()))
            if "id" in variant:
                del variant["id"]
            writer.writerow(variant)

        QApplication.instance().clipboard().setText(output.getvalue())
        output.close()

    def _open_default_link(self, index: QModelIndex):

        current_variant = self.model.variant(index.row())
        full_variant = sql.get_one_variant(self.conn, current_variant["id"])
        # Update variant with annotation data visible in the view ...
        full_variant.update(current_variant)
        #  get default link
        link = [i for i in self._get_links() if i["is_default"] is True]
        if not link:
            return

        link = link[0]

        url = self._create_url(link["url"], full_variant)
        if url:
            self._open_url(url, link["is_browser"])

    def on_double_clicked(self, index: QModelIndex):
        """ 
        React on double clicked
        TODO : duplicate code with ContextMenu Event ! Need to refactor a bit 
        """

        self._open_default_link(index)


class VariantViewWidget(plugin.PluginWidget):
    """Contains the view of query with several controller"""

    variant_clicked = Signal(dict)
    LOCATION = plugin.CENTRAL_LOCATION

    ENABLE = True

    def __init__(self, parent=None):
        super().__init__(parent)

        # Create 2 Panes
        self.splitter = QSplitter(Qt.Horizontal)
        self.main_right_pane = VariantView(parent=self)

        # No popup menu on this one
        self.groupby_left_pane = VariantView(parent=self, show_popup_menu=False)
        self.groupby_left_pane.hide()
        # Force selection of first item after refresh
        self.groupby_left_pane.row_to_be_selected = 0

        self.splitter.addWidget(self.groupby_left_pane)
        self.splitter.addWidget(self.main_right_pane)

        # Make resizable TODO : ugly ... Make it nicer
        # def _resize1_section(l, o, n):
        #     self.groupby_left_pane.view.horizontalHeader().resizeSection(l, n)

        # def _resize2_section(l, o, n):
        #     self.main_right_pane.view.horizontalHeader().resizeSection(l, n)

        # self.main_right_pane.view.horizontalHeader().sectionResized.connect(_resize1_section)
        # self.groupby_left_pane.view.horizontalHeader().sectionResized.connect(
        #     _resize2_section
        # )

        # self.groupby_left_pane.view.setHorizontalHeader(self.main_right_pane.view.horizontalHeader())

        # Top toolbar
        self.top_bar = QToolBar()
        # PS: Actions with QAction::LowPriority will not show the text labels
        self.top_bar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)

        # checkable group action
        self.groupby_action = self.top_bar.addAction(
            FIcon(0xF14E0), self.tr("Group by"), self.on_group_changed
        )
        self.groupby_action.setToolTip(
            self.tr("Group variants according to choosen columns")
        )
        self.groupby_action.setCheckable(True)
        self.groupby_action.setChecked(False)

        # groupbylist
        self.groupbylist_action = self.top_bar.addAction("chr,pos,ref")
        self.groupbylist_action.setVisible(False)
        self.groupbylist_action.triggered.connect(self._show_group_dialog)

        self.top_bar.addSeparator()
        # Save selection
        self.save_action = self.top_bar.addAction(
            FIcon(0xF0F87), self.tr("Save selection"), self.on_save_selection
        )
        self.save_action.setToolTip(
            self.tr("Save the current variants into a new selection")
        )
        # self.save_action.setPriority(QAction.LowPriority)

        # Refresh UI button
        action = self.top_bar.addAction(
            FIcon(0xF0450), self.tr("Refresh"), self.on_refresh
        )
        action.setToolTip(self.tr("Refresh the current list of variants"))
        # action.setPriority(QAction.LowPriority)

        # Interrupt current query
        action = self.top_bar.addAction(
            FIcon(0xF04DB), self.tr("Stop"), self.on_interrupt
        )
        action.setToolTip(self.tr("Stop current query"))

        #  edit action
        # TODO : move into view

        # Formatter tools
        self.top_bar.addSeparator()
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.top_bar.addWidget(spacer)

        # Add formatters to combobox
        self.formatter_combo = QComboBox()
        self.add_available_formatters()

        # Error handling
        self.log_edit = QLabel()
        self.log_edit.setMaximumHeight(30)
        self.log_edit.setStyleSheet(
            "QWidget{{background-color:'{}'; color:'{}'}}".format(
                style.WARNING_BACKGROUND_COLOR, style.WARNING_TEXT_COLOR
            )
        )
        self.log_edit.hide()
        self.log_edit.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)

        # Setup layout
        main_layout = QVBoxLayout()
        main_layout.addWidget(self.top_bar)
        main_layout.addWidget(self.splitter)
        main_layout.addWidget(self.log_edit)

        self.setLayout(main_layout)

        # Make connection
        self.main_right_pane.view.selectionModel().currentRowChanged.connect(
            lambda x, _: self.on_variant_clicked(x)
        )

        self.groupby_left_pane.view.selectionModel().currentRowChanged.connect(
            lambda x, _: self.on_variant_clicked(x)
        )
        self.groupby_left_pane.no_variant.connect(self.on_no_variant)
        # Connect errors from async runnables
        self.main_right_pane.error_raised.connect(self.set_message)
        self.groupby_left_pane.error_raised.connect(self.set_message)

        # Default group
        self.last_group = ["chr"]

        # Save fields between group/ungroup
        self.save_fields = list()
        self.save_filters = list()

    def add_available_formatters(self):
        """Populate the formatters

        Also recall previously selected formatters via config file.
        Default formatter is "SeqoneFormatter".
        """
        # Get previously selected formatter

        settings = QSettings()
        settings.beginGroup("variant_view")
        formatter_name = settings.value("formatter", "CutestyleFormatter")

        # Add formatters to combobox, a click on it will instantiate the class
        selected_formatter_index = 0
        for index, obj in enumerate(formatter.find_formatters()):
            self.formatter_combo.addItem(FIcon(0xF03D8), obj.DISPLAY_NAME, obj)
            if obj.__name__ == formatter_name:
                selected_formatter_index = index

        self.top_bar.addWidget(self.formatter_combo)
        self.formatter_combo.currentTextChanged.connect(self.on_formatter_changed)
        self.formatter_combo.setToolTip(self.tr("Change current style"))

        # Set the previously used/default formatter
        self.formatter_combo.setCurrentIndex(selected_formatter_index)

    def on_formatter_changed(self):
        """Activate the selected formatter and save it in settings

        Called when the current formatter is modified
        """

        settings = QSettings()
        settings.beginGroup("variant_view")

        formatter_class = self.formatter_combo.currentData()
        self.main_right_pane.set_formatter(formatter_class())
        self.groupby_left_pane.set_formatter(formatter_class())
        # Save formatter setting
        settings.setValue("formatter", formatter_class.__name__)

    def on_open_project(self, conn):
        """Overrided from PluginWidget"""
        self.conn = conn
        # Set connections of models
        self.main_right_pane.conn = self.conn
        self.groupby_left_pane.conn = self.conn

        self.on_refresh()

    def on_refresh(self):
        """Overrided from PluginWidget"""
        # Save default data with current query attributes
        # See load(), we use this attr to restore fields after grouping
        self.save_fields = self.mainwindow.state.fields
        self.save_filters = self.mainwindow.state.filters

        self.main_right_pane.fields = self.mainwindow.state.fields
        self.main_right_pane.source = self.mainwindow.state.source
        self.main_right_pane.filters = self.mainwindow.state.filters
        # Set current group_by to left pane
        self._set_groups(self.mainwindow.state.group_by)

        # Set formatter
        formatter_class = self.formatter_combo.currentData()

        #        formatter_class = next(formatter.find_formatters())
        self.main_right_pane.set_formatter(formatter_class())
        self.groupby_left_pane.set_formatter(formatter_class())

        # Clear only the variant cache ! Because user can edit data
        self.main_right_pane.model.clear_variant_cache()
        self.groupby_left_pane.model.clear_variant_cache()

        # Load ui
        self.load(reset_page=True)

    def on_interrupt(self):
        self.main_right_pane.model.interrupt()
        self.groupby_left_pane.model.interrupt()

    def on_save_selection(self):
        """Triggered on 'save_selection' button

        - This slot creates a new selection (aka source) from the current state.
        - `source_editor` plugin will be refreshed.
        """

        selection_name, accept = QInputDialog.getText(
            self,
            self.tr("Create a new selection"),
            self.tr("Name of the new selection:"),
        )

        if not accept or not selection_name:
            return

        try:
            result = cmd.create_cmd(
                self.conn,
                selection_name,
                source=self.mainwindow.state.source,
                filters=self.mainwindow.state.filters,
                count=self.main_right_pane.model.total,
            )
            if result:
                self.mainwindow.refresh_plugin("source_editor")
        except Exception as e:
            LOGGER.exception(e)

            # Name already used
            QMessageBox.critical(
                self,
                self.tr("Error while creating selection"),
                self.tr("Error while creating selection '%s'; Name is already used")
                % selection_name,
            )
            # Retry
            self.on_save_selection()

    def on_no_variant(self):
        """Slot called when left pane hasn't any variant to display

        Clear right pane (nothing to select in it)
        """
        self.main_right_pane.model.clear()

    def _is_grouped(self) -> bool:
        """Return grouped mode status of the view"""
        # print("is grouped ?")
        # print("left", self.groupby_left_pane.model.group_by)
        # print("right", self.main_right_pane.model.group_by)
        return self.groupby_left_pane.group_by != []

    def load(self, reset_page=False):
        """Load all views

        Called by on_refresh, on_group_changed, and _show_group_dialog
        Display/hide groupby_left_pane on user demand.
        """
        is_grouped = self._is_grouped()
        # Left pane and groupbylist are visible in group mode
        self.groupby_left_pane.setVisible(is_grouped)
        self.groupbylist_action.setVisible(is_grouped)

        self.groupby_action.blockSignals(True)
        self.groupby_action.setChecked(is_grouped)
        self.groupby_action.blockSignals(False)

        # Hide potential errors
        self.log_edit.hide()

        if is_grouped:
            # Ungrouped => grouped or already grouped
            # Groupby fields become left pane fields
            self.groupby_left_pane.fields = self.groupby_left_pane.group_by
            # Prune right fields with left fields => avoid redundancy of information
            self.main_right_pane.fields = [
                field
                for field in self.save_fields
                if field not in self.groupby_left_pane.group_by
            ]
            self.groupby_left_pane.filters = self.save_filters
            self.groupby_left_pane.source = self.main_right_pane.source

            # Refresh models
            # right pane: Useless, except if we modify fields like above
            # But refreshing left pane triggers right pane refreshing
            # So => do not do it here
            # self.main_right_pane.load()
            self.groupby_left_pane.load(reset_page)
        else:
            # Grouped => ungrouped
            # Restore fields
            self.main_right_pane.fields = self.save_fields
            # Restore filters
            self.main_right_pane.filters = self.save_filters
            # Refresh model
            self.main_right_pane.load(reset_page)
            # print("saved right:", self.save_fields)

    def on_group_changed(self):
        """Set group by fields when group by button is clicked"""
        is_checked = self.groupby_action.isChecked()
        is_grouped = self._is_grouped()
        if is_checked and not is_grouped:
            # Group it
            self.groupby_action.setIcon(FIcon(0xF14E1))
            self.groupby_action.setToolTip(self.tr("Ungroup variants"))
            # Recall previous/default group
            self._set_groups(self.last_group)
        else:
            # Ungroup it
            self.groupby_action.setIcon(FIcon(0xF14E0))
            self.groupby_action.setToolTip(
                self.tr("Group variants according to choosen columns")
            )
            # Save current group
            self.last_group = self.groupby_left_pane.group_by
        if not is_checked:
            # Reset to default group (chr set in _show_group_dialog)
            self._set_groups([])

        self.load()
        self._refresh_vql_editor()

    def on_variant_clicked(self, index: QModelIndex):
        """React on variant clicked

        Args:
            index (QModelIndex): index into item models derived from
                QAbstractItemModel. U used by item views, delegates, and
                selection models to locate an item in the model.
        """

        if index.model() == self.groupby_left_pane.view.model():
            # Variant clicked on left pane => refresh the right pane
            variant = self.groupby_left_pane.model.variant(index.row())

            if self._is_grouped():
                # Restore fields
                self.groupby_left_pane.source = self.main_right_pane.source

                # Forge a special filter to display the current variant
                and_list = [
                    {
                        "field": field,
                        "operator": "=",
                        "value": variant[fields_to_vql(field)],
                    }
                    for field in self.groupby_left_pane.group_by
                ]

                if self.save_filters:
                    # Build a new filter dict/tree based on the current one
                    filters = copy.deepcopy(self.save_filters)

                    if "AND" in self.save_filters:
                        # {'AND': [...]}
                        # => Just add our additional constraints joined by AND
                        filters["AND"] += and_list
                    elif "OR" in self.save_filters:
                        # {'OR': [...]}
                        # => Insert the previous filter as a child of a AND constraint
                        # Ours are mandatory
                        # {'AND': [{'OR': [...]}, ...]}
                        and_list.append(filters)
                        filters = {"AND": and_list}
                else:
                    # New filter => {'AND': [...]}
                    filters = {"AND": and_list}

                LOGGER.debug("Filters for right pane for grouped variants %s", filters)

                self.main_right_pane.filters = filters

                # Update right pane only
                self.main_right_pane.load(reset_page=True)

        if index.model() == self.main_right_pane.view.model():
            # Variant clicked on right pane
            variant = self.main_right_pane.model.variant(index.row())

        # Refresh the current variant of mainwindow and plugins
        self.mainwindow.state.current_variant = variant
        # Request a refresh of the variant_info plugin
        self.mainwindow.refresh_plugin("variant_info")

    def _show_group_dialog(self):
        """Show a dialog window to select group fields"""
        dialog = QDialog(self)

        view = QListWidget()
        box = QVBoxLayout()
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        box.addWidget(view)
        box.addWidget(buttons)
        buttons.accepted.connect(dialog.accept)
        buttons.rejected.connect(dialog.reject)

        # Populate the list of fields with their check status
        for field in self.save_fields:
            item = QListWidgetItem()
            item.setText(fields_to_vql(field))
            item.setData(Qt.UserRole, field)
            if field in self.groupby_left_pane.group_by:
                item.setCheckState(Qt.Checked)
            else:
                item.setCheckState(Qt.Unchecked)

            view.addItem(item)
        dialog.setLayout(box)

        if dialog.exec_() == QDialog.Accepted:

            selected_fields = []
            for i in range(view.count()):
                item = view.item(i)
                if item.checkState() == Qt.Checked:
                    selected_fields.append(item.data(Qt.UserRole))

            # if no group force chr
            if not selected_fields:
                selected_fields = ["chr"]

            # Set current group_by to left pane
            self._set_groups(selected_fields)
            # Reload ui
            self.load()
            self._refresh_vql_editor()

    def _set_groups(self, grouped_fields):
        """Set fields on groupby_left_pane and refresh text on groupby action"""
        self.groupby_left_pane.group_by = grouped_fields
        self.groupbylist_action.setText(
            ",".join([fields_to_vql(field) for field in grouped_fields])
        )

    def _refresh_vql_editor(self):
        """Force refresh of vql_editor if loaded"""
        if "vql_editor" in self.mainwindow.plugins:
            self.mainwindow.state.group_by = self.groupby_left_pane.group_by
            # Request a refresh of the vql_editor plugin
            self.mainwindow.refresh_plugin("vql_editor")

    def copy(self):
        """Copy the selected variant(s) into the clipboard

        See Also: VariantView.copy_to_clipboard
        """
        self.main_right_pane.copy_to_clipboard()

    def select_all(self):
        """Select all variants in the view

        See Also: VariantView.select_all
        """
        self.main_right_pane.select_all()

    def set_message(self, message: str):
        """Show message error at the bottom of the view

        Args:
            message (str): Error message
        """

        if self.log_edit.isHidden():
            self.log_edit.show()

        icon_64 = FIcon(0xF0027, style.WARNING_TEXT_COLOR).to_base64(18, 18)

        self.log_edit.setText(
            """<div height=100%>
            <img src="data:image/png;base64,{}" align="left"/>
             <span> {} </span>
            </div>""".format(
                icon_64, message
            )
        )


if __name__ == "__main__":
    import sys
    from PySide2.QtWidgets import QApplication
    from cutevariant.core.importer import import_file, import_reader
    from cutevariant.core.reader import FakeReader, VcfReader
    from cutevariant.core import sql

    app = QApplication(sys.argv)

    m = QStringListModel()
    m.setStringList(["salut", "sach"])
    w = LoadingTableView()
    w.setModel(m)
    w.show()

    # conn = sql.get_sql_connection(":memory:")
    # reader = VcfReader(
    #     open("/home/sacha/Dev/cutevariant/examples/test.snpeff.vcf"), "snpeff"
    # )
    # import_reader(conn, reader)

    # w = VariantViewWidget()
    # w.conn = conn
    # w.main_right_pane.model.conn = conn
    # w.main_right_pane.load()
    # # w.main_view.model.group_by = ["chr","pos","ref","alt"]
    # # w.on_refresh()

    # w.show()

    app.exec_()



"""Functions to build a complex SQL Select statement to query variant.

In the most of cases, you will only use build_sql_query function.

Examples::

    conn = sqlite3.Connection("::memory::")
    query = build_sql_query(["chr","pos"])
    conn.execute(query)


A variant query can be understood from the following 3 points of view.

The SQL statements:
~~~~~~~~~~~~~~~~~~~
This is the raw sqlite query::

    "SELECT `variants`.`chr`, `variants`.`pos` FROM `variants` WHERE `variants`.`ref` == 'A' "

The VQL statements:
~~~~~~~~~~~~~~~~~~~
This is a domain specific language. Check the vql module for more information::

    "SELECT chr, pos FROM variants WHERE ref == 'A'"

The Python statements:
~~~~~~~~~~~~~~~~~~~~~~
The representation of the query as a dictionnary::

    query = {
        "fields": ["chr","pos"],
        "source": "variants",
        "filters": {
        "AND":
            [
                {"name":"ref", "operator":"==","value":"A"}
            ]
        }
    }

The following module contains several functions to make easy conversion from
Python statements to VQL or SQL statements .

In the most of the case, you only need to use :

- :meth:`build_vql_query`: to create a VQL query from python statements
- :meth:`build_sql_query`: to create a SQL query from python statements
"""
# Standard imports
import sqlite3
import re
from functools import lru_cache
from ast import literal_eval

# Custom imports
from cutevariant.core import sql
from cutevariant.commons import logger

LOGGER = logger()

# TODO : can be move somewhere else ? In common ?
# Function names used in VQL
# sample["boby"].gt
GENOTYPE_FUNC_NAME = "sample"
# WORDSET["truc"]
WORDSET_FUNC_NAME = "WORDSET"


def filters_to_flat(filters: dict):
    """Recursive function to convert the filter hierarchical dictionnary into a list of fields

    Args:
        filters (dict): a nested tree of condition. @See example

    Returns:
        (list): all fields are now inside a a list

    .. todo:: Move to vql ?

    Examples::

        filters = {
            'AND': [
                {'field': 'ref', 'operator': '=', 'value': "A"},
                {'field': 'alt', 'operator': '=', 'value': "C"}
            ]
        }

        filters = _flatten_filter(filters)

        filters is now:
        [
            {'field': 'ref', 'operator': '=', 'value': "A"},
            {'field': 'alt', 'operator': '=', 'value': "C"}
        ]
    """

    def recursive_generator(filters):
        if isinstance(filters, dict) and len(filters) == 3:
            # length = 3 to skip AND/OR levels
            yield filters

        if isinstance(filters, dict):
            for i in filters:
                yield from recursive_generator(filters[i])

        if isinstance(filters, list):
            for i in filters:
                yield from recursive_generator(i)

    return list(recursive_generator(filters))


def field_function_to_sql(field_function: tuple, use_as=False):
    """Convert VQL function to a a jointure field name

    Examples:

        >>> # which correspond to genotype(boby).GT in VQL
        >>> field = ("genotype", "boby", "gt")
        >>> field_function_to_sql(field)
        "`genotype_boby`.`gt`"
    """
    func_name, arg_name, field_name = field_function

    if use_as:
        # THIS IS INSANE... SQLITE DOESNT RETURN ALIAS NAME WITH SQUARE BRACKET....
        # I HAVE TO replace [] by () and go back after...
        # TODO : Change VQL Syntax from [] to () would be a good alternative
        # See : https://stackoverflow.com/questions/41538952/issue-cursor-description-never-returns-square-bracket-in-column-name-python-2-7-sqlite3-alias
        alias_name = fields_to_vql(field_function).replace("[", "(").replace("]", ")")
        suffix = ' AS "{}"'.format(alias_name)
    else:
        suffix = ""

    if field_name:
        return f"`{func_name}_{arg_name}`.`{field_name}`" + suffix
    return f"`{func_name}_{arg_name}`" + suffix


def wordset_data_to_sql(wordset_expr: tuple):
    """Get the SQL version of a Wordset expression (`(WORDSET', 'boby')`)

    Wordset function is used in VQL to filter fields within a set of words.

    Example:

        .. code-block:: sql

            SELECT ... WHERE gene IN WORDSET('boby')
            -- will be replaced by:
            SELECT ... WHERE gene IN (SELECT value FROM sets WHERE name = 'boby')

        We return only the sub SELECT statement here::

            >>> wordset_data_to_sql(("WORDSET", "boby"))
            "SELECT value FROM sets WHERE name = 'boby'"

    Args:
        wordset_expr (tuple): Tuple of 2 items: First one is "WORDSET",
            second one is the name of the queried wordset.
    Returns:
        (str): Query statement
    """
    func_name, arg_name = wordset_expr
    assert func_name == WORDSET_FUNC_NAME
    return f"(SELECT value FROM wordsets WHERE name = '{arg_name}')"


def wordset_data_to_vql(wordset_expr: tuple):
    """Get the VQL version of a Wordset expression (`(WORDSET', 'boby')`)

    Example:

        >>> wordset_data_to_vql(("WORDSET", "boby"))
        "WORDSET['boby']"

    Args:
        wordset_expr (tuple): Tuple of 2 items: First one is "WORDSET",
            second one is the name of the queried wordset.
    Returns:
        (str): Query statement
    """
    return "{}['{}']".format(*wordset_expr)


def fields_to_vql(field) -> str:
    """Return field as VQL syntax

    This is used to convert tuple field and create a VQL query

    Examples:
        >>> field = ("sample", "boby", "gt")
        >>> field_to_vql(field)
        "sample['boby'].gt"

    Args:
        field(str or tuple): a Field

    Returns:
        str: fields for vql query
    """
    if isinstance(field, tuple):
        if field[0] == GENOTYPE_FUNC_NAME and len(field) == 3:
            return f"{field[0]}['{field[1]}'].{field[2]}"
    # str
    return field


def fields_to_sql(field, default_tables={}, use_as=False) -> str:
    """Return field as SQL syntax

    Args:
        field (str or tuple): Column name from a table
        default_tables (dict, optional): association between field name and table origin

    Returns:
        str: Sql field

    Examples:

        >>> default_tables = {"variants.chr": "variants", "alt": "variants"}
        >>> fields_to_sql("chr", default_tables=default_tables)
        "`variants`.`chr`"
        >>> fields_to_sql("gene", default_tables=default_tables)
        "`annotations`.`gene`"
        >>> fields_to_sql("coucou", default_tables=default_tables)
        "`coucou`"
    """
    if isinstance(field, tuple):
        # If it is "genotype.name.truc then it is field function"
        # ("genotype", "boby", "gt") => `genotype_boby`.GT
        return field_function_to_sql(field, use_as)

    # extract variants.chr => (variants, chr) => `variants`.`chr`
    match = re.match(r"^(\w+)\.(\w+)", field)

    if match:
        table = match[1]
        field = match[2]
    else:
        if field in default_tables.keys():
            # Set the table
            table = default_tables[field]
            # Samples fields are managed differently from JOIN... Skip it
            if table == "samples":
                return f"`{field}`"

        else:
            # Unknown table, just return the field
            return f"`{field}`"

    return f"`{table}`.`{field}`"


def filters_to_sql(filters, default_tables={}):
    """Return filters as SQL syntax

    Args:
        filters (dict): Nested tree of conditions
        default_tables (dict, optional): Association between field names and tables

    Returns:
        str: SQL WHERE expression

    Examples:
        >>> filters_to_sql({"AND": [("pos", ">", 34), ("af", "=", 10)})
        "`variants`.`pos` > 34 AND `variants`.`af` = 10"

    Note:
        There is a recursive function inside to parse the nested tree of conditions
    """

    def is_field(node):
        return len(node) == 3

    def recursive(node):
        if not node:
            return ""

        if is_field(node):
            # print("Node to SQL", node)
            field = fields_to_sql(node["field"], default_tables)
            value = node["value"]
            operator = node["operator"].upper()

            if operator == "~":
                operator = "REGEXP"

            if operator in ("IN", "NOT IN") and isinstance(value, str):
                # node: {'field': 'ref', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"}
                # wanted: "ref IN ('A', 'T', 'G', 'C')"
                # Try to cast string to tuple
                try:
                    temp_val = literal_eval(value)
                    value = temp_val
                except ValueError:
                    pass

            elif (operator == "IS" or operator == "IS NOT") and value == "NULL":
                value = f"NULL"

            elif isinstance(value, str):
                # enclose string with quotes
                value = f"'{value}'"

            if isinstance(value, tuple):
                # If value is a WORDSET["salut"] aka ('WORDSET', 'salut')
                # => Get SQL statement
                if value[0] == WORDSET_FUNC_NAME:
                    value = wordset_data_to_sql(value)
                elif len(value) == 1:
                    # Remove trailing comma in tuple with 1 element ("xxx",)
                    value = str(value).replace(",", "")

            # Strings must be space separated because of operators (IN, etc.)
            return "%s %s %s" % (field, operator, value)

        # If node is not a field: 1 key only: the logical operator
        logic_op = list(node.keys())[0]
        g = (recursive(child) for child in node[logic_op])
        out = [sql_filter for sql_filter in g if sql_filter]
        # print("to SQL", out, "LOGIC", logic_op)
        # Recursive call for each field in the list associated to the
        # logical operator.
        # node:
        # {'AND': [
        #   {'field': 'ref', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"},
        #   {'field': 'alt', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"}
        # ]}
        # Wanted: "ref IN ('A', 'T', 'G', 'C') AND alt IN ('A', 'T', 'G', 'C')"
        # to SQL ["ref IN ('A', 'T', 'G', 'C')", "alt IN ('A', 'T', 'G', 'C')"]
        if not out:
            return ""
        if len(out) == 1:
            return out[0]
        return "(" + f" {logic_op} ".join(out) + ")"

    return recursive(filters)


def filters_to_vql(filters):
    """Return filters as VQL syntax

    Args:
        filters (dict): Nested tree of conditions

    Returns:
        str: VQL WHERE expression
    """

    def is_field(node):
        return len(node) == 3

    def recursive(node):
        if not node:
            return ""

        if is_field(node):
            # print("Node to VQL", node)
            field = fields_to_vql(node["field"])
            value = node["value"]
            operator = node["operator"].upper()

            if operator in ("IN", "NOT IN") and isinstance(value, str):
                # node: {'field': 'ref', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"}
                # wanted: "ref IN ('A', 'T', 'G', 'C')"
                # Try to cast string to tuple
                try:
                    temp_val = literal_eval(value)
                    value = temp_val
                except ValueError:
                    pass

            elif (operator == "IS" or operator == "IS NOT") and value == "NULL":
                value = f"NULL"

            elif isinstance(value, str):
                value = f"'{value}'"

            if isinstance(value, tuple):
                # If value is ('WORDSET', 'salut') aka WORDSET["salut"]
                # => Get VQL statement
                if len(value) == 2 and value[0] == WORDSET_FUNC_NAME:
                    value = wordset_data_to_vql(value)
                elif len(value) == 1:
                    # Remove trailing comma in tuple with 1 element ("xxx",)
                    value = str(value).replace(",", "")

            # Strings must be space separated because of operators (IN, etc.)
            return "%s %s %s" % (field, operator, value)

        # If node is not a field: 1 key only: the logical operator
        logic_op = list(node.keys())[0]
        g = (recursive(child) for child in node[logic_op])
        out = [sql_filter for sql_filter in g if sql_filter]
        # print("to VQL", out, "LOGIC", logic_op)
        # {'field': 'chr', 'operator': 'IN', 'value': (10.0, 11.0)}
        # to VQL ['chr IN (10.0, 11.0)'] LOGIC AND
        if not out:
            return ""
        if len(out) == 1:
            return out[0]
        return "(" + f" {logic_op} ".join(out) + ")"

    # Do not return first level of enclosing parenthesis
    vql_filters = recursive(filters)
    if vql_filters.startswith("("):
        return vql_filters[1:-1]
    return vql_filters


def build_vql_query(fields, source="variants", filters={}, group_by=[], having={}):
    """Build VQL SELECT query

    Args:
        fields (list): List of fields
        source (str): source of the virtual table ( see: selection )
        filters (dict): nested condition tree
        group_by (list/None): list of field you want to group
    """
    query = "SELECT " + ",".join([fields_to_vql(i) for i in fields]) + " FROM " + source
    if filters:
        where_clause = filters_to_vql(filters)
        if where_clause:
            query += " WHERE " + where_clause

    if group_by:
        query += " GROUP BY " + ",".join((fields_to_vql(i) for i in group_by))

        if having:
            operator = having["op"]
            value = having["value"]
            query += f" HAVING count {operator} {value}"

    return query


def build_sql_query(
    fields,
    source="variants",
    filters={},
    order_by=None,
    order_desc=True,
    limit=50,
    offset=0,
    group_by=[],
    having={},  # {"op":">", "value": 3  }
    default_tables={},
    samples_ids={},
    **kwargs,
):
    """Build SQL SELECT query

    Args:
        fields (list): List of fields
        source (str): source of the virtual table ( see: selection )
        filters (dict): nested condition tree
        order_by (str/None): Order by field;
            If None, order_desc is not required.
        order_desc (bool): Descending or Ascending order
        limit (int/None): limit record count;
            If None, offset is not required.
        offset (int): record count per page
        group_by (list/None): list of field you want to group
        default_tables (dict): association map between fields and sql table origin
        samples_ids (dict): association map between samples name and id
    """
    # Create fields
    sql_fields = ["`variants`.`id`"] + [
        fields_to_sql(col, default_tables, use_as=True) for col in fields if col != "id"
    ]

    # if group_by:
    #     sql_fields.insert(1, "COUNT() as 'count'")

    sql_query = f"SELECT DISTINCT {','.join(sql_fields)} "

    # #Add child count if grouped
    # if grouped:
    #     sql_query += ", COUNT(*) as `children`"

    # Add source table
    sql_query += "FROM variants"

    # Extract fields from filters
    fields_in_filters = {i["field"] for i in filters_to_flat(filters)}

    # Loop over fields and check is annotations is required
    annotation_fields = {i for i, v in default_tables.items() if v == "annotations"}

    need_join_annotations = False
    for col in set(sql_fields) | fields_in_filters:
        # Example of field:
        # '`annotations`.`gene`'
        if "annotations" in col or col in annotation_fields:
            need_join_annotations = True
            break

    if need_join_annotations:
        sql_query += " LEFT JOIN annotations ON annotations.variant_id = variants.id"

    # Add Join Selection
    # TODO: set variants as global variables
    if source != "variants":
        sql_query += (
            " INNER JOIN selection_has_variant sv ON sv.variant_id = variants.id "
            f"INNER JOIN selections s ON s.id = sv.selection_id AND s.name = '{source}'"
        )

    # Add Join Samples
    ## detect if fields contains function like (genotype,boby,gt) and save boby
    all_fields = set(fields_in_filters)
    all_fields.update(fields)

    samples = []
    for col in all_fields:
        # if column looks like  "genotype.tumor.gt"
        if isinstance(col, tuple):
            if col[0] == GENOTYPE_FUNC_NAME:
                sample_name = col[1]
                samples.append(sample_name)

    # make samples uniques
    samples = set(samples)

    ## Create Sample Join
    for sample_name in samples:
        # Optimisation ?
        # sample_id = self.cache_samples_ids[sample_name]
        if sample_name in samples_ids:
            sample_id = samples_ids[sample_name]
            sql_query += f""" INNER JOIN sample_has_variant `{GENOTYPE_FUNC_NAME}_{sample_name}` ON `{GENOTYPE_FUNC_NAME}_{sample_name}`.variant_id = variants.id AND `{GENOTYPE_FUNC_NAME}_{sample_name}`.sample_id = {sample_id}"""

    # Add Where Clause
    if filters:
        where_clause = filters_to_sql(filters, default_tables)
        if where_clause:
            sql_query += " WHERE " + where_clause

    # Add Group By
    if group_by:
        sql_query += " GROUP BY " + ",".join(
            [fields_to_sql(g, default_tables, use_as=False) for g in group_by]
        )
        if having:
            operator = having["op"]
            val = having["value"]
            sql_query += f" HAVING count {operator} {val}"

    # Add Order By
    if order_by:
        # TODO : sqlite escape field with quote
        orientation = "DESC" if order_desc else "ASC"
        order_by = fields_to_sql(order_by, default_tables)
        sql_query += f" ORDER BY {order_by} {orientation}"

    if limit:
        sql_query += f" LIMIT {limit} OFFSET {offset}"

    return sql_query


def build_full_sql_query(
    conn: sqlite3.Connection,
    fields=["chr", "pos", "ref", "alt"],
    source="variants",
    filters=dict(),
    order_by=None,
    order_desc=True,
    group_by=[],
    having={},  # {"op":">", "value": 3  }
    limit=50,
    offset=0,
    **kwargs,
):
    """Build a complete SQL SELECT statement according to the data loaded from DB

    You don't have to give the association map between fields and sql table origin
    nor the association map between samples name and id.
    In exchange SQL connection is mandatory.

    Args:
        conn (sqlite3.Connection): SQL connection
        fields (list): List of fields
        source (str): source of the virtual table ( see: selection )
        filters (dict): nested condition tree
        order_by (str/None): Order by field;
            If None, order_desc is not required.
        order_desc (bool): Descending or Ascending order
        limit (int/None): limit record count;
            If None, offset is not required.
        offset (int): record count per page
        group_by (list/None): list of field you want to group
    """
    # Used cached data
    default_tables, sample_ids = get_default_tables_and_sample_ids(conn)
    # LOGGER.debug(get_default_tables_and_sample_ids.cache_info())

    query = build_sql_query(
        fields=fields,
        source=source,
        filters=filters,
        order_by=order_by,
        order_desc=order_desc,
        limit=limit,
        offset=offset,
        group_by=group_by,
        having=having,
        default_tables=default_tables,
        samples_ids=sample_ids,
        **kwargs,
    )
    return query


@lru_cache()
def get_default_tables_and_sample_ids(conn):
    """Handy function to cache default_tables and sample_ids from database

    This function is used for every queries built in :meth:`build_full_sql_query`

    Warnings:
        Do not forget to clear this cache when samples are added in DB via
        a PED file for example.
    """
    # Get {'favorite': 'variants', 'comment': 'variants', impact': 'annotations', ...}
    default_tables = {i["name"]: i["category"] for i in sql.get_fields(conn)}
    # Get {'NORMAL': 1, 'TUMOR': 2}
    sample_ids = {i["name"]: i["id"] for i in sql.get_samples(conn)}

    return default_tables, sample_ids



# Standard imports
from pkg_resources import parse_version

# Qt imports
from PySide2 import __version__ as pyside_version
from PySide2.QtCore import Qt
from PySide2.QtGui import QKeySequence, QIcon
from PySide2.QtWidgets import (
    QTextEdit,
    QDialog,
    QApplication,
    QVBoxLayout,
    QToolBar,
    QPlainTextEdit,
    QSplitter,
    QDialogButtonBox,
)

# Custom imports
from cutevariant.gui.ficon import FIcon
from cutevariant import commons as cm

LOGGER = cm.logger()


class MarkdownEditor(QDialog):
    """Markdown editor used to add comments on variants

    On PySide2 5.14+, comments can be edited in Markdown and previewed.

    """

    def __init__(self, parent=None, default_text=""):
        """Init Markdown editor

        :param default_text: Populate source edit with current comment in db.
        :type default_text: <str>
        """
        super().__init__(parent)

        self.setWindowTitle("Cutevariant - " + self.tr("Comment editor"))
        self.setWindowIcon(QIcon(cm.DIR_ICONS + "app.png"))

        main_vlayout = QVBoxLayout()
        # main_vlayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_vlayout)

        # Setup edit view
        self.rich_edit = QTextEdit()  # Rich text result
        self.source_edit = QPlainTextEdit()  # Markdown content
        self.source_edit.setPlainText(default_text)

        vlayout = QVBoxLayout()
        vlayout.setSpacing(1)
        vlayout.setContentsMargins(0, 0, 0, 0)
        main_vlayout.addLayout(vlayout)

        self.splitter = QSplitter(Qt.Horizontal)
        self.splitter.addWidget(self.source_edit)

        if parse_version(pyside_version) >= parse_version("5.14"):
            # RichText in Markdown is supported starting PySide 5.14
            self.splitter.addWidget(self.rich_edit)
            # self.rich_edit.setStyleSheet(
            #     "QWidget {background-color:'lightgray'}")
            self.rich_edit.setAcceptRichText(True)
            self.rich_edit.setAutoFormatting(QTextEdit.AutoAll)
            self.rich_edit.setReadOnly(True)

            # Update preview with Markdown content
            self.source_edit.textChanged.connect(self.update_rich_text)
            # Auto refresh rich text content now
            self.update_rich_text()

        # Setup toolbar
        self.toolbar = QToolBar()
        self.act_undo = self.toolbar.addAction(self.tr("undo"), self.source_edit.undo)
        self.act_undo.setIcon(FIcon(0xF054C))
        self.act_undo.setShortcut(QKeySequence.Undo)

        self.act_redo = self.toolbar.addAction(self.tr("redo"), self.source_edit.redo)
        self.act_redo.setIcon(FIcon(0xF044E))
        self.act_redo.setShortcut(QKeySequence.Redo)

        self.act_bold = self.toolbar.addAction(
            self.tr("bold"), lambda: self.infix("**")
        )
        self.act_bold.setIcon(FIcon(0xF0264))
        self.act_bold.setShortcut(QKeySequence("CTRL+B"))

        self.act_italic = self.toolbar.addAction(
            self.tr("italic"), lambda: self.infix("*")
        )
        self.act_italic.setIcon(FIcon(0xF0277))
        self.act_italic.setShortcut(QKeySequence("CTRL+I"))

        self.act_heading = self.toolbar.addAction(
            self.tr("insert title"), lambda: self.heading()
        )
        self.act_heading.setShortcut(QKeySequence("CTRL+H"))
        self.act_heading.setIcon(FIcon(0xF0274))

        self.act_unorder_list = self.toolbar.addAction(
            self.tr("insert list item"), self.unorder_list
        )
        self.act_unorder_list.setIcon(FIcon(0xF0279))

        self.act_quote = self.toolbar.addAction(self.tr("quote"), self.quote)
        self.act_quote.setIcon(FIcon(0xF027E))

        self.act_link = self.toolbar.addAction(self.tr("insert link"), self.link)
        self.act_link.setIcon(FIcon(0xF0339))

        vlayout.addWidget(self.toolbar)
        vlayout.addWidget(self.splitter)

        # Add buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Save | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        main_vlayout.addWidget(buttons)

    def update_rich_text(self):
        """Update preview with Markdown content

        .. warning:: Depends PySide Qt5.14
        """
        self.rich_edit.setMarkdown(self.source_edit.toPlainText())

    def infix(self, prefix: str, suffix=None):
        """Add tags before and after the selected text"""
        if suffix is None:
            suffix = prefix

        cursor = self.source_edit.textCursor()

        if not cursor.hasSelection():
            cursor.insertText(f"{prefix}Text{suffix}")
            return

        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        # must be made in 1 step to avoid flooding the history (do/undo)
        text = self.source_edit.toPlainText()
        cursor.insertText(prefix + text[start:end] + suffix)

    def heading(self):
        """Add header tag before selected text"""
        cursor = self.source_edit.textCursor()

        if not cursor.hasSelection():
            cursor.insertText("# Title")
            return

        start = cursor.selectionStart()
        cursor.setPosition(start)
        cursor.insertText("# ")

    def unorder_list(self):
        """Add list tag before selected text"""
        cursor = self.source_edit.textCursor()

        if not cursor.hasSelection():
            cursor.insertText("- List item")
            return

        start = cursor.selectionStart()
        cursor.setPosition(start)
        cursor.insertText("- ")

    def quote(self):
        """Quote the selected text"""
        cursor = self.source_edit.textCursor()

        if not cursor.hasSelection():
            cursor.insertText("> text")
            return

        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        # must be made in 1 step to avoid flooding the history (do/undo)
        text = self.source_edit.toPlainText()
        cursor.insertText("\n> %s\n" % text[start:end])

    def link(self):
        """Format selected text into URL link"""
        cursor = self.source_edit.textCursor()

        if not cursor.hasSelection():
            cursor.insertText("[text](url)")
            return

        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        # must be made in 1 step to avoid flooding the history (do/undo)
        text = self.source_edit.toPlainText()
        cursor.insertText("[text](%s)" % text[start:end])

    def toPlainText(self):
        """Get source text from the current comment

        Used to save the comment in the database by variant_view plugin.
        """
        return self.source_edit.toPlainText()


if __name__ == "__main__":
    import sys
    from cutevariant.gui.ficon import setFontPath
    import cutevariant.commons as cm

    app = QApplication(sys.argv)

    setFontPath(cm.FONT_FILE)

    edit = MarkdownEditor()

    edit.show()

    app.exec_()



"""Design pattern "COMMANDS" to execute VQL query .

Each VQL statement corresponds to a <name>_cmd() function.
You can use `execute(conn, vql)` or `execute_all(conn, vql)` to run a specific
VQL query.

Each command returns a JSON array with a success status code or with the
expected result.

Command module is usefull for CLI and for running VQL scripts.

Example:

    >>> conn = sqlite.Connection("project.db")
    >>> for variant in execute(conn, "SELECT chr, pos FROM variants"):
    ...     print(variant)
    >>> # How many variants ?
    >>> print(execute(conn, "COUNT FROM variants"))
"""
# Standard imports
import sqlite3
import os
import functools


# Custom imports
from cutevariant.core.querybuilder import build_sql_query, build_full_sql_query
from cutevariant.core import sql, vql
from cutevariant.commons import logger
from cutevariant.core.reader import BedReader

LOGGER = logger()


def select_cmd(
    conn: sqlite3.Connection,
    fields=("chr", "pos", "ref", "alt"),
    source="variants",
    filters={},
    order_by=None,
    order_desc=True,
    group_by=[],
    having={},  # {"op":">", "value": 3  }
    limit=50,
    offset=0,
    **kwargs,
):
    """Select query Command

    This following VQL command:
        `SELECT chr,pos FROM variants WHERE pos > 3`
    will execute :
        `select_cmd(conn, ["chr", "pos"], variants", {"AND": [{"pos","=",3}]}))`

    Args:
        conn (sqlite3.Connection): sqlite3 connection
        fields (list, optional): list of fields
        filters (dict, optional): nested tree of condition
        source (str, optional): virtual table source
        order_by (list, optional): order by field name
        order_desc (bool, optional): Descending or Ascending Order
        limit (int, optional): record count
        offset (int, optional): record count per page

    Yields:
        variants (dict)
    """
    query = build_full_sql_query(
        conn,
        fields=fields,
        source=source,
        filters=filters,
        order_by=order_by,
        order_desc=order_desc,
        limit=limit,
        offset=offset,
        group_by=group_by,
        having=having,
        **kwargs,
    )
    LOGGER.debug("command:select_cmd:: %s", query)
    for i in conn.execute(query):
        # THIS IS INSANE... SQLITE DOESNT RETURN ALIAS NAME WITH SQUARE BRACKET....
        # I HAVE TO replace [] by () and go back after...
        # TODO : Change VQL Syntax from [] to () would be a good alternative
        # @See QUERYBUILDER
        # See : https://stackoverflow.com/questions/41538952/issue-cursor-description-never-returns-square-bracket-in-column-name-python-2-7-sqlite3-alias
        yield {k.replace("(", "[").replace(")", "]"): v for k, v in dict(i).items()}


def count_cmd(
    conn: sqlite3.Connection,
    fields=["chr", "pos", "ref", "alt"],
    source="variants",
    filters={},
    group_by=[],
    having={},
    **kwargs,
):
    """Count command

    This following VQL command:
        `COUNT FROM variants WHERE pos > 3`
    will execute :
        `count_cmd(conn, "variants", {"AND": [{"pos","=",3}]}))`

    Args:
        conn (sqlite3.Connection): sqlite3 connection
        source (str, optional): virtual source table
        filters (dict, optional): nested tree of condition

    Returns:
        dict: Count of variants wgstith "count" as a key
    """
    # See #177: Check if fields has annotations
    # If an annotation field is selected, the variant count stored in the selection
    # table (count without taking account of annotations) is different.
    # This leads to a fault in the pagination hiding the latest variants if
    # more than 50 must be displayed.

    variants_fields = set(
        field["name"] for field in sql.get_field_by_category(conn, "variants")
    )

    if set(fields).issubset(variants_fields) and not filters and not group_by:
        # All fields are in variants table
        # Returned stored cache variant
        LOGGER.debug("command:count_cmd:: cached from selections table")
        return {
            "count": conn.execute(
                f"SELECT count FROM selections WHERE name = '{source}'"
            ).fetchone()[0]
        }

    query = build_full_sql_query(
        conn,
        fields=fields,
        source=source,
        filters=filters,
        limit=None,
        offset=None,
        order_by=None,
        group_by=group_by,
        having=having,
        **kwargs,
    )

    # THIS IS INSANE... SQLITE DOESNT RETURN ALIAS NAME WITH SQUARE BRACKET....
    # I HAVE TO replace [] by () and go back after...
    # TODO : Change VQL Syntax from [] to () would be a good alternative
    # @See QUERYBUILDER
    # See : https://stackoverflow.com/questions/41538952/issue-cursor-description-never-returns-square-bracket-in-column-name-python-2-7-sqlite3-alias
    LOGGER.debug("command:count_cmd:: %s", query)
    return {"count": sql.count_query(conn, query)}


def drop_cmd(conn: sqlite3.Connection, feature: str, name: str, **kwargs):
    """Drop selection or set from database

    This following VQL commands::

        DROP selections boby
        DROP wordsets mygene

    will execute::

        drop_cmd(conn, "selections", "boby")
        drop_cmd(conn, "wordsets", "boby")

    Args:
        conn (sqlite3.Connection): sqlite connection
        feature (str): selections or wordsets (Names of the SQL tables).
            Lower case features are also accepted.
        name (str): name of the selection or name of the wordset

    Returns:
        dict: {"success": <boolean>}; True if deletion is ok, False otherwise.

    Raises:
        vql.VQLSyntaxError
    """
    accept_features = ("selections", "wordsets")

    # Cast to lower case
    feature = feature.lower()

    if feature not in accept_features:
        raise vql.VQLSyntaxError(f"{feature} doesn't exists")

    affected_lines = sql.delete_by_name(conn, name, table_name=feature)
    return {"success": (affected_lines > 0)}


def create_cmd(
    conn: sqlite3.Connection,
    target: str,
    source="variants",
    filters=dict(),
    count=None,
    **kwargs,
):
    """Create a selection from the given source, filtered by filters

    This following VQL command:
        `CREATE boby FROM variants WHERE pos > 3`
    will execute :
        `create_cmd(conn, "boby", "variants", {"AND":[{"pos",">",3}]})`

    Args:
        conn (sqlite3.Connection): sqlite3 connection
        target (str): target selection name
        source (str): source selection name
        filters (dict): filters used to select the variants from source
        count (int): precomputed variant count

    Returns:
        dict: {"id": selection_id} if lines have been inserted,
            or empty dict in case of error
    """
    if target is None:
        return {}

    sql_query = build_full_sql_query(
        conn, fields=["id"], source=source, filters=filters, limit=None, **kwargs,
    )

    LOGGER.debug("command:create_cmd:: %s", sql_query)
    selection_id = sql.create_selection_from_sql(
        conn, sql_query, target, count=count, from_selection=False
    )
    return dict() if selection_id is None else {"id": selection_id}


def set_cmd(
    conn: sqlite3.Connection, target: str, first: str, second: str, operator, **kwargs
):
    """Perform set operation like intersection, union and difference between two table selection

    This following VQL command:
        `CREATE boby = raymond & charles`
    will execute :
        `set_cmd(conn, "boby", "raymond", "charles", "&")`

    Args:
        conn (sqlite3.Connection): sqlite3 connection
        target (str): table selection target
        first (str): first selection in operation
        second (str): second selection in operation
        operator (str): + (union), - (difference), & (intersection) Set operators

    Returns:
        dict: {"id": selection_id} if lines have been inserted,
            or empty dict in case of error

    Examples:
        {"id": 2}: 2 lines inserted
    """
    if target is None or first is None or second is None or operator is None:
        return {}

    query_first = build_sql_query(["id"], first, limit=None)
    query_second = build_sql_query(["id"], second, limit=None)

    func_query = {
        "|": sql.union_variants,
        "-": sql.subtract_variants,
        "&": sql.intersect_variants,
    }

    sql_query = func_query[operator](query_first, query_second)
    LOGGER.debug("command:set_cmd:: %s", sql_query)

    selection_id = sql.create_selection_from_sql(
        conn, sql_query, target, from_selection=False
    )
    return dict() if selection_id is None else {"id": selection_id}


def bed_cmd(conn: sqlite3.Connection, path: str, target: str, source: str, **kwargs):
    """Create a new selection from a bed file

    This following VQL command:
        `CREATE boby FROM variants INTERSECT "path/to/file.bed"`
    will execute :
        `bed_cmd(conn, "path/to/file.bed", "boby", "source")`

    Args:
        conn (sqlite3.Connection): sqlite3 connection
        path (str): path to bedfile ( a 3 columns files with chr, start, end )
        target (str): target selection table
        source (str): source selection table

    Returns:
        dict: {"id": id of last line inserted} if lines have been inserted,
            or empty dict in case of error

    Raises:
        vql.VQLSyntaxError
    """
    if not os.path.isfile(path):
        raise vql.VQLSyntaxError(f"{path} doesn't exists")

    # bed_intervals: chrom, start, end, name, etc. keys in each interval
    # see also cutevariant/core/reader/bedreader.py
    selection_id = sql.create_selection_from_bed(conn, source, target, BedReader(path))
    return dict() if selection_id is None else {"id": selection_id}


def show_cmd(conn: sqlite3.Connection, feature: str, **kwargs):
    """Show command display information from a SHOW query

    This following VQL command:
        `SHOW variants`
    will execute :
        `show_cmd(conn, "variants")`

    Args:
        conn (sqlite3.Connection): sqlite3 connection
        feature (str): Requested feature type of items (Name of the SQL table);
            Lower case features are also accepted.
        can be: `"selections", "fields", "samples", "wordsets"`

    Yields:
        (generator[dict]): Items according to requested feature.

    Raises:
        vql.VQLSyntaxError
    """
    accepted_features = {
        "selections": sql.get_selections,
        "fields": sql.get_fields,
        "samples": sql.get_samples,
        "wordsets": sql.get_wordsets,
    }

    # Cast in lower case
    feature = feature.lower()

    if feature not in accepted_features:
        raise vql.VQLSyntaxError(f"option {feature} doesn't exists")

    for item in accepted_features[feature](conn):
        yield item


def import_cmd(conn: sqlite3.Connection, feature=str, name=str, path=str, **kwargs):
    """Import command for wordsets only

    This following VQL command:
        `IMPORT WORDSETS "gene.txt" AS boby`
    will execute :
        `import_cmd(conn, "WORDSETS", "gene.txt")`

    Args:
        conn (sqlite3.Connection): sqlite3 connection
        feature (str): "WORDSETS" (Name of the SQL table)
        name (str): name of the set
        path (str): a filepath

    Returns:
        dict: `{success: <boolean>}`

    Raises:
        vql.VQLSyntaxError
    """
    accepted_features = ("wordsets",)

    if feature.lower() not in accepted_features:
        raise vql.VQLSyntaxError(f"option {feature} doesn't exists")

    if not os.path.isfile(path):
        raise vql.VQLSyntaxError(f"{path} doesn't exists")

    affected_rows = sql.import_wordset_from_file(conn, name, path)
    return {"success": (affected_rows is not None)}


def create_command_from_obj(conn, vql_obj: dict):
    """Create command function from vql object according to its "cmd" key

    cmd authorized:

        - select_cmd
        - count_cmd
        - drop_cmd
        - create_cmd
        - set_cmd
        - bed_cmd
        - show_cmd
        - import_cmd

    Warning:
        Use :meth:`execute` instead, that is a wrapper to this function.

    Examples:
        If you want to create a select_cmd function, pass a vql_obj like this one:

        .. code-block:: python

            {
                "cmd": "select_cmd",
                "fields": ["chr", "pos"],
                "source": "variants",
                "filters": {},
            }

    Args:
        conn (sqlite3.Connection): sqlite3.connection
        vql_obj (dict): A VQL object with requested commands at "cmd" key:
            `select_cmd, create_cmd, set_cmd, bed_cmd, show_cmd, import_cmd,
            drop_cmd, count_cmd`.
            A VQL object is dictionary returned by vql.parse.

    Returns:
        (function): Function object wrapping the given vql_object.
    """
    command = vql_obj["cmd"]
    if command in globals():
        return functools.partial(globals()[command], conn, **vql_obj)
    raise vql.VQLSyntaxError(f"cmd {command} doesn't exists")


def execute(conn: sqlite3.Connection, vql_source: str):
    """Execute a vql query

    Examples:
        >>> for variant in execute(conn,"SELECT chr from variants"):
        >>>     print(variant)

    Args:
        conn (sqlite3.Connection): sqlite3 connection
        vql_source (str): a VQL query

    Returns:
        dict: Return command output as a dict
    """
    # Convert VQL string into VQL object
    vql_obj = vql.parse_one_vql(vql_source)
    # Convert VQL object into wrapped SQL query
    cmd = create_command_from_obj(conn, vql_obj)
    return cmd()


def execute_all(conn: sqlite3.Connection, vql_source: str):
    """Execute a vql script

    Examples:
        >>> execute_all(
        ...     "CREATE boby FROM variants; CREATE raymon FROM variants;"
        ...     "CREATE charles = boby - variants; COUNT(charles)"
        ... )

    Args:
        conn (sqlite3.Connection): sqlite3 connection
        vql_source (str): a VQL query

    Yields:
        (generator[dict]): Yield command outputs as a dicts
    """
    for vql_obj in vql.parse_vql(vql_source):
        cmd = create_command_from_obj(conn, vql_obj)
        yield cmd()


# class CommandGraph(object):
#     def __init__(self, conn):
#         super().__init__()
#         self.conn = conn
#         self.graph = nx.DiGraph()
#         self.graph.add_node("variants")

#     def add_command(self, command: Command):

#         if type(command) == CreateCommand:
#             self.graph.add_node(command.target)
#             self.graph.add_edge(command.source, command.target)

#         if type(command) == SelectCommand:
#             self.graph.add_node("Select")
#             self.graph.add_edge(command.source, "Select")

#         if type(command) == SetCommand:
#             self.graph.add_node(command.target)
#             self.graph.add_edge(command.first, command.target)
#             self.graph.add_edge(command.second, command.target)

#     def set_source(self, source):
#         self.graph.clear()
#         for vql_obj in vql.parse_vql(source):
#             cmd = create_command_from_vql_objet(self.conn, vql_obj)
#             self.add_command(cmd)



# Qt imports
from PySide2.QtCore import Qt, QAbstractTableModel
from PySide2.QtWidgets import QToolBar, QVBoxLayout, QApplication

from cutevariant.gui import style, plugin, FIcon
from cutevariant.core.querybuilder import build_vql_query
from cutevariant.commons import logger

LOGGER = logger()


class HistoryModel(QAbstractTableModel):

    HEADERS = ["time", "count", "query"]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.records = []

    def rowCount(self, parent: QModelIndex) -> int:
        """ override """
        return len(self.records)

    def columnCount(self, parent: QModelIndex) -> int:
        """ override """
        return 3

    def data(self, index: QModelIndex, role):
        """ override """

        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            if index.column() == 0:
                return self.records[index.row()][0].toString("hh:mm:ss")
            if index.column() == 1:
                return str(self.records[index.row()][1])
            if index.column() == 2:
                return self.records[index.row()][2]
        if role == Qt.ToolTipRole:
            return self.records[index.row()][2]

        return None

    def headerData(self, section: int, orientation: Qt.Orientation, role):
        """ override """

        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.HEADERS[section]

        return None

    def add_record(self, query: str, count: int):
        """Add a record into the model
        
        Args:
            query (str): A VQL query
            count (int): the total count of variant returns by the VQL query
        """

        now = QDateTime().currentDateTime()

        self.beginInsertRows(QModelIndex(), 0, 0)
        self.records.insert(0, (now, count, query))
        self.endInsertRows()

    def clear_records(self):
        """Clear records from models
        """
        self.beginResetModel()
        self.records.clear()
        self.endResetModel()

    def get_record(self, index: QModelIndex):
        """ Return record corresponding to the model index """
        return self.records[index.row()]


class VqlHistoryWidget(plugin.PluginWidget):
    """Exposed class to manage VQL/SQL queries from the mainwindow"""

    LOCATION = plugin.FOOTER_LOCATION
    ENABLE = True
    REFRESH_ONLY_VISIBLE = False

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle(self.tr("VQL Editor"))

        # Create model / view
        self.view = QTableView()
        self.model = HistoryModel()
        self.view.setModel(self.model)
        self.view.setAlternatingRowColors(True)
        self.view.horizontalHeader().setStretchLastSection(True)
        self.view.verticalHeader().hide()
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)

        #  Getting the variant count is not easy in this way...
        # Because the variant count is computed asynchronously from the variant_view
        #  TODO : need to find a way ! I hide the column for now
        self.view.hideColumn(1)

        self.view.doubleClicked.connect(self.on_double_clicked)
        #  Create toolbar
        self.toolbar = QToolBar()
        self.toolbar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.toolbar.addAction(
            FIcon(0xF0413), self.tr("Clear"), self.model.clear_records
        )

        # Create layout
        main_layout = QVBoxLayout()
        main_layout.setSpacing(0)
        main_layout.addWidget(self.toolbar)
        main_layout.addWidget(self.view)
        main_layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(main_layout)

    def on_register(self, mainwindow):
        """ override """
        pass

    def on_open_project(self, conn):
        """ override """
        pass

    def on_close(self):
        """ override """
        pass

    def on_refresh(self):
        """ override """

        vql_query = build_vql_query(
            self.mainwindow.state.fields,
            self.mainwindow.state.source,
            self.mainwindow.state.filters,
            self.mainwindow.state.group_by,
            self.mainwindow.state.having,
        )

        # TODO : Get the variant count
        # Get the total count from variant_view is not easy because it is asynchrone...
        #  So it is hidden for now ...

        self.model.add_record(vql_query, 0)

    def on_double_clicked(self, index: QModelIndex):
        """triggered when history record is clicked 
        
        Args:
            index (QModelIndex): index 
        """
        _, _, query = self.model.get_record(index)
        parsed_query = next(vql.parse_vql(query))
        print(parsed_query)

        self.mainwindow.state.fields = parsed_query["fields"]
        self.mainwindow.state.source = parsed_query["source"]
        self.mainwindow.state.filters = parsed_query["filters"]
        self.mainwindow.state.group_by = parsed_query["group_by"]
        self.mainwindow.state.having = parsed_query["having"]

        self.mainwindow.refresh_plugins(sender=self)


if __name__ == "__main__":

    import sys
    import sqlite3

    app = QApplication(sys.argv)

    conn = sqlite3.connect("/home/sacha/Dev/cutevariant/examples/test.db")

    view = VqlHistoryWidget()
    view.show()

    app.exec_()



# Standard imports
from logging.handlers import RotatingFileHandler
import logging
import datetime as dt
import re
import json
import tempfile
import os
from pkg_resources import resource_filename

# Misc
MAX_RECENT_PROJECTS = 5
MIN_COMPLETION_LETTERS = 1
DEFAULT_SELECTION_NAME = "variants"
# version from which database files are supported (included)
MIN_AUTHORIZED_DB_VERSION = "0.2.0"


# ACMG Classification
CLASSIFICATION = {
    0: "Unclassified",
    1: "Benin",
    2: "Likely benin",
    3: "Variant of uncertain significance",
    4: "Likely pathogen",
    5: "Pathogen",
}

CLASSIFICATION_ICONS = {
    0: 0xF03A1,
    1: 0xF03A4,
    2: 0xF03A7,
    3: 0xF03AA,
    4: 0xF03AD,
    5: 0xF03B1,
}

# Genotypes
GENOTYPE_ICONS = {-1: 0xF10D3, 0: 0xF0766, 1: 0xF0AA1, 2: 0xF0AA5}

GENOTYPE_DESC = {
    -1: "Unknown genotype",
    0: "Homozygous wild",
    1: "Heterozygous",
    2: "Homozygous muted",
}

# Paths
DIR_LOGS = tempfile.gettempdir() + "/"

DIR_ASSETS = resource_filename(__name__, "assets/")  # current package name
DIR_TRANSLATIONS = DIR_ASSETS + "i18n/"
DIR_FONTS = DIR_ASSETS + "fonts/"
DIR_ICONS = DIR_ASSETS + "icons/"
DIR_STYLES = DIR_ASSETS + "styles/"

BASIC_STYLE = "Bright"
FONT_FILE = DIR_FONTS + "materialdesignicons-webfont.ttf"

# Websites and variant query
WEBSITES_URLS = {
    "GenCards - The human gene database": "https://www.genecards.org/cgi-bin/carddisp.pl?gene={gene}",
    "Varsome - Genes": "https://varsome.com/gene/{gene}",
    "Varsome - Variants": "https://varsome.com/variant/hg19/{chr}-{pos}-{ref}-{alt}",
}

REPORT_BUG_URL = "https://github.com/labsquare/cutevariant/issues/new"
WIKI_URL = "https://github.com/labsquare/cutevariant/wiki"

# Logging
LOGGER_NAME = "cutevariant"
LOG_LEVEL = "DEBUG"
LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "notset": logging.NOTSET,
}

################################################################################


def logger(name=LOGGER_NAME, logfilename=None):
    """Return logger of given name, without initialize it.

    Equivalent of logging.getLogger() call.
    """
    logger = logging.getLogger(name)
    FORMAT = "%(levelname)s: [%(filename)s:%(lineno)s:%(funcName)s()] %(message)s"
    logging.basicConfig(format=FORMAT)
    return logger


_logger = logging.getLogger(LOGGER_NAME)
_logger.setLevel(LOG_LEVEL)

# log file
formatter = logging.Formatter("%(asctime)s :: %(levelname)s :: %(message)s")
file_handler = RotatingFileHandler(
    DIR_LOGS
    + LOGGER_NAME
    + "_"
    + dt.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    + ".log",
    "a",
    100_000_000,
    1,
)
file_handler.setLevel(LOG_LEVEL)
file_handler.setFormatter(formatter)
_logger.addHandler(file_handler)

# terminal log
# stream_handler = logging.StreamHandler()
# formatter = logging.Formatter("%(levelname)s: %(message)s")
# stream_handler.setFormatter(formatter)
# stream_handler.setLevel(LOG_LEVEL)
# _logger.addHandler(stream_handler)


def log_level(level):
    """Set terminal/file log level to given one.
    .. note:: Don't forget the propagation system of messages:
        From logger to handlers. Handlers receive log messages only if
        the main logger doesn't filter them.
    """
    # Main logger
    _logger.setLevel(level.upper())
    # Handlers
    [
        handler.setLevel(level.upper())
        for handler in _logger.handlers
        if handler.__class__
        in (logging.StreamHandler, logging.handlers.RotatingFileHandler)
    ]


def is_gz_file(filepath):
    """Return a boolean according to the compression state of the file"""
    with open(filepath, "rb") as test_f:
        return test_f.read(3) == b"\x1f\x8b\x08"


def get_uncompressed_size(filepath):
    """Get the size of the given compressed file
    This size is stored in the last 4 bytes of the file.
    """
    with open(filepath, "rb") as file_obj:
        file_obj.seek(-4, 2)
        return int.from_bytes(file_obj.read(4), byteorder="little")


def bytes_to_readable(size) -> str:
    """return human readable size from bytes
    
    Args:
        size (int): size in bytes
    
    Returns:
        str: readable size
    """
    out = ""
    for count in ["Bytes", "KB", "MB", "GB"]:
        if size > -1024.0 and size < 1024.0:
            return "%3.1f%s" % (size, count)
        size /= 1024.0
    return "%3.1f%s" % (size, "TB")


def snake_to_camel(name: str) -> str:
    """Convert snake_case name to CamelCase name

    Args:
        name (str): a snake string like : query_view

    Returns:
        str: a camel string like: QueryView
    """
    return "".join([i.capitalize() for i in name.split("_")])


def camel_to_snake(name: str) -> str:
    """Convert CamelCase to snake_case name

    Args:
        name (str): a snake string like : QueryView

    Returns:
        str: a camel string like: query_view
    """
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def is_json_file(filename):

    if not os.path.exists(filename):
        return False

    with open(filename) as file:
        try:
            json.load(file)
        except Exception as e:
            return False
        finally:
            return True



# Standard imports
import tempfile

# Qt imports
from PySide2.QtCore import (
    Qt,
    QAbstractTableModel,
    QAbstractItemModel,
    QModelIndex,
    Property,
    Signal,
)

from PySide2.QtWidgets import (
    QTableView,
    QItemDelegate,
    QWidget,
    QStyleOptionViewItem,
    QComboBox,
    QAbstractItemView,
)

# Custom imports
from cutevariant.core.reader import PedReader
from cutevariant.core.writer import PedWriter


class PedModel(QAbstractTableModel):
    """
    Attributes:

        samples_data (list[list]): List of samples; Each sample is composed of
            the following terms:

            `[family_id, individual_id, father_id, mother_id, sex, genotype]`

    """

    SEX_MAP = {"1": "Male", "2": "Female", "0": ""}
    PHENOTYPE_MAP = {"1": "Unaffected", "2": "Affected", "0": ""}
    UNKNOWN_VALUES = ["", "NA", "0"]

    def __init__(self):
        super().__init__()

        self.samples_data = []
        self.headers = (
            "Family",
            "Sample",
            "Father_id",
            "Mother_id",
            "Sex",
            "Phenotype",
        )

    def rowCount(self, index=QModelIndex()) -> int:
        """ override """
        return len(self.samples_data)

    def columnCount(self, index=QModelIndex()) -> int:
        """ override """
        if index == QModelIndex():
            return len(self.headers)

        return 0

    def get_data_list(self, column: int):
        """Get unique list of items at the given column of all samples

        Notes:
            `samples_data` is composed of the following terms:
            `[family_id, individual_id, father_id, mother_id, sex, genotype]`

        Examples:
            If column 1 is given, we return a list of unique individual_ids.
        """
        return list({sample[column] for sample in self.samples_data})

    def clear(self):
        self.beginResetModel()
        self.samples_data.clear()
        self.endResetModel()

    def from_pedfile(self, filename: str):
        """Fill model with NEW samples from PED file"""
        samples = dict()
        self.beginResetModel()
        self.samples_data.clear()
        self.samples_data = list(PedReader(filename, samples))
        self.endResetModel()

    def to_pedfile(self, filename: str):
        """Export the model to a tabulated PED file

        Notes:
            Replace None or empty strings to 0 (unknown PED ID)
        """
        with open(filename, "w") as file:
            writer = PedWriter(file)
            writer.save_from_list(self.samples_data)

    def set_samples(self, samples: list):
        """Fill model with NEW samples

        Example:
            samples = [family_id, individual_id, father_id, mother_id, sex, genotype]
        """

        self.beginResetModel()
        self.samples_data.clear()
        self.samples_data = list(samples)
        self.endResetModel()

    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        """Overrided

        Notes:
            Default values of different sample fields are empty strings
        """
        if not index.isValid():
            return

        if role == Qt.DisplayRole or role == Qt.EditRole:
            value = self.samples_data[index.row()][index.column()]

            if index.column() == 2 or index.column() == 3:  # father_id, mother_id
                return value if value not in self.UNKNOWN_VALUES else ""

            if index.column() == 4:  # Sex
                return self.SEX_MAP.get(str(value), "")

            if index.column() == 5:  # Phenotype
                return self.PHENOTYPE_MAP.get(str(value), "")

            return value

    def setData(self, index: QModelIndex, value, role=Qt.EditRole):
        """ overrided """

        if not index.isValid():
            return

        if role == Qt.EditRole:
            self.samples_data[index.row()][index.column()] = value
            return True

        return False

    def headerData(
        self, section: int, orientation: Qt.Orientation, role: Qt.DisplayRole
    ):
        """ overrided """
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                return self.headers[section]

    def flags(self, index: QModelIndex):
        """ overrided """
        if not index.isValid():
            return

        if index.column() > 1:
            # Family ids & Individual ids are NOT editable (we must fit with the VCF file)
            return Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled

        return Qt.ItemIsSelectable | Qt.ItemIsEnabled


class PedDelegate(QItemDelegate):
    """Allow the way items of data are rendered and edited to be customized

    Signals:
        parthenogenesis_detected(str): Emit message about the sample that has
            the same father and mother ids.
    """

    parthenogenesis_detected = Signal(str)

    def __init__(self):
        super().__init__()
        # Keep the rows of erroneous samples (same father/mother ids)
        self.erroneous_samples = set()

    def createEditor(
        self, parent: QWidget, option: QStyleOptionViewItem, index: QModelIndex
    ):
        """Return editor widget for columns of PedView

        Notes:
            Family ids & Individual ids are NOT editable (we must fit with the VCF file)
        """
        # PS: index.model refer to SampleModel

        if index.column() < 2:
            # Family ids & Individual ids are NOT editable (we must fit with the VCF file)
            return

        widget = QComboBox(parent)
        if index.column() == 2 or index.column() == 3:
            # Forge a list of available individual ids except the current one
            # Get individual_id of the current row/sample
            current_individual_id = index.model().samples_data[index.row()][1]
            # Remove current individual_id from propositions
            individual_ids = set(index.model().get_data_list(1))
            individual_ids.remove(current_individual_id)
            # father_id or mother_id columns
            widget.addItem("", "0")
            for individual_id in individual_ids:
                widget.addItem(individual_id, individual_id)

            return widget

        if index.column() == 4:
            # Sex column
            widget.addItem("Male", "1")
            widget.addItem("Female", "2")
            widget.addItem("", "0")
            return widget

        if index.column() == 5:
            # Genotype column
            widget.addItem("Unaffected", "1")
            widget.addItem("Affected", "2")
            widget.addItem("", "0")
            return widget

        return super().createEditor(parent, option, index)

    def setModelData(
        self, editor: QWidget, model: QAbstractItemModel, index: QModelIndex
    ):
        """Set the data for the item at the given index in the model to the
        contents of the given editor.
        """
        if isinstance(editor, QComboBox):
            model.setData(index, editor.currentData())

            if index.column() == 2 or index.column() == 3:
                row = index.row()
                model = index.model()
                # Detect parthenogenesis: same mother and father
                father_id = model.samples_data[row][2]
                # Only for not unknown parents
                if father_id != "0" and father_id == model.samples_data[row][3]:
                    self.erroneous_samples.add(row)
                elif row in self.erroneous_samples:
                    # Reset interface
                    self.parthenogenesis_detected.emit("")
                    self.erroneous_samples.remove(row)

            for row in self.erroneous_samples:
                self.parthenogenesis_detected.emit(
                    self.tr("<b>Same father and mother for sample '{}'</b>").format(
                        model.samples_data[row][1]
                    )
                )
            return

        # Basic text not editable
        return super().setModelData(editor, model, index)


class PedView(QTableView):
    """View to display a PED file content

    Expose properties binding for QWizardPage.registerFields

    Attributes:
        samples(list): List of PED fields
        pedfile(str): PED filepath
    """

    message = Signal(str)

    def __init__(self):
        super().__init__()
        self.model = PedModel()
        self.delegate = PedDelegate()
        self.setModel(self.model)
        self.horizontalHeader().setStretchLastSection(True)
        self.setAlternatingRowColors(True)
        self.verticalHeader().hide()
        self.setItemDelegate(self.delegate)
        self.setEditTriggers(
            QAbstractItemView.CurrentChanged | QAbstractItemView.DoubleClicked
        )
        self.delegate.parthenogenesis_detected.connect(self.message)
        # PED file for the model
        self.outfile = None

    def clear(self):
        self.model.clear()

    @property
    def samples(self):
        """Get samples (list of PED fields)"""
        return self.model.samples_data

    @samples.setter
    def samples(self, samples):
        """Set samples

        Args:
            samples(list): PED fields
        """
        self.model.set_samples(samples)

    @Property(str)  # Qt property for QWizardPage.registerFields
    def pedfile(self):
        """Return the filepath of the PED file associated to the current model"""
        if not self.samples:
            return

        if not self.outfile:
            # Same file but reused at each call
            self.outfile = tempfile.mkstemp(suffix=".ped", text=True)[1]
        # Export the PED data
        self.model.to_pedfile(self.outfile)
        return self.outfile



from PySide2.QtCore import QSettings
from PySide2.QtNetwork import QNetworkAccessManager, QNetworkProxy

from cutevariant.commons import logger


LOGGER = logger()

PROXY_TYPES = {
    "No Proxy": QNetworkProxy.NoProxy,
    "Default": QNetworkProxy.DefaultProxy,
    "Sock5": QNetworkProxy.Socks5Proxy,
    "Http": QNetworkProxy.HttpProxy,
}


def get_network_manager():

    # Create network access manager
    network_manager = QNetworkAccessManager()

    # Get proxy settings data
    settings = QSettings("labsquare", "cutevariant")
    settings.beginGroup("proxy")
    p_type_index = settings.value("type", 0)
    p_host = settings.value("host")
    p_port = settings.value("port", 80)
    p_username = settings.value("username")
    p_password = settings.value("password")
    settings.endGroup()

    try:
        p_type = list(PROXY_TYPES.values())[int(p_type_index)]

        if p_port:
            p_port = int(p_port)

        if p_type is not QNetworkProxy.NoProxy:
            proxy = QNetworkProxy(p_type, p_host, p_port, p_username, p_password)
            network_manager.setProxy(proxy)
    except Exception as e:
        LOGGER.exception(e)
        pass

    return network_manager



# Standard imports
import os
import copy
import time

# Qt imports
from PySide2.QtWidgets import *
from PySide2.QtCore import QThread, Signal, QDir, QSettings, QFile, Slot, Qt
from PySide2.QtGui import QIcon, QStandardItem, QStandardItemModel

# Custom imports
from cutevariant.core.importer import async_import_file
from cutevariant.core import get_sql_connection
import cutevariant.commons as cm
from cutevariant.core.readerfactory import detect_vcf_annotation, create_reader
from cutevariant.core.reader import PedReader
from cutevariant.gui.model_view import PedView

LOGGER = cm.logger()


class ProjectPage(QWizardPage):
    """Page: Creation of a new project
    We ask the reference genome, the name and the path of the future project file.
    """

    def __init__(self):
        super().__init__()

        self.setTitle(self.tr("Project creation"))
        self.setSubTitle(
            self.tr("This wizard will guide you to create a cutevariant project.")
        )

        # Reload last directory used
        app_settings = QSettings()
        self.last_directory = app_settings.value("last_directory", QDir.homePath())

        self.project_name_edit = QLineEdit()
        self.project_path_edit = QLineEdit()
        self.project_path_edit.setText(self.last_directory)
        self.browse_button = QPushButton(self.tr("Browse"))
        self.reference = QComboBox()

        # Unused for now
        self.reference.hide()

        self.reference.addItem("hg19")
        self.registerField("project_name", self.project_name_edit, "text")
        self.registerField("project_path", self.project_path_edit, "text")
        self.registerField("reference", self.reference, "currentText")

        v_layout = QFormLayout()

        browse_layout = QHBoxLayout()
        browse_layout.addWidget(self.project_path_edit)
        browse_layout.addWidget(self.browse_button)
        browse_layout.setContentsMargins(0, 0, 0, 0)

        v_layout.addRow(self.tr("Reference genom"), self.reference)
        v_layout.addRow(self.tr("Project Name"), self.project_name_edit)
        v_layout.addRow(self.tr("Create in"), browse_layout)

        self.setLayout(v_layout)

        self.browse_button.clicked.connect(self._browse)
        self.project_path_edit.textChanged.connect(self.completeChanged)
        self.project_name_edit.textChanged.connect(self.completeChanged)

    @Slot()
    def _browse(self):
        """Open a dialog box to set the directory where the project will be saved"""
        path = QFileDialog.getExistingDirectory(
            self, self.tr("Select a path for the project"), self.last_directory
        )
        if path:
            self.project_path_edit.setText(path)

    def initializePage(self):
        """Overridden: Prepare the page just before it is shown"""
        # Adjust the focus of project name field
        self.project_name_edit.setFocus()

    def isComplete(self):
        """Conditions to unlock next button"""
        return (
            True
            if (
                QDir(self.project_path_edit.text()).exists()
                and self.project_path_edit.text()
                and self.project_name_edit.text()
            )
            else False
        )


class FilePage(QWizardPage):
    """Page: Open the file containing variant data
    We ask the path of the data file
    """

    def __init__(self):
        super().__init__()

        self.setTitle(self.tr("Select a file"))
        self.setSubTitle(self.tr("Supported file are vcf, vcf.gz, vep.txt."))

        self.file_path_edit = QLineEdit()

        self.anotation_detect_label = QLabel()
        self.button = QPushButton(self.tr("Browse"))
        h_layout = QHBoxLayout()

        h_layout.addWidget(self.file_path_edit)
        h_layout.addWidget(self.button)

        self.v_layout = QVBoxLayout()
        self.v_layout.addLayout(h_layout)
        self.v_layout.addWidget(self.anotation_detect_label)
        self.setLayout(self.v_layout)

        self.button.clicked.connect(self._browse)
        self.file_path_edit.textChanged.connect(self.completeChanged)

        self.registerField("filename", self.file_path_edit, "text")
        # annotation ? should be an option or not ?
        self.registerField("annotation", self.anotation_detect_label, "text")

    @Slot()
    def _browse(self):
        """Open a dialog box to set the data file

        The file is opened and we show an alert if annotations are detected.
        """
        # Reload last directory used
        app_settings = QSettings()
        last_directory = app_settings.value("last_directory", QDir.homePath())

        filepath, filetype = QFileDialog.getOpenFileName(
            self,
            self.tr("Open a file"),
            last_directory,
            self.tr("VCF file (*.vcf *.vcf.gz);; CSV file (*.csv *.tsv *.txt)"),
        )

        if filepath:
            # Display and save directory
            self.file_path_edit.setText(filepath)
            app_settings.setValue("last_directory", os.path.dirname(filepath))

            if "vcf" in filepath:
                # TODO: detect annotations on other tyes of files...
                annotation_type = detect_vcf_annotation(filepath)
                if annotation_type:
                    text = self.tr("<b>%s annotations detected!</b>") % annotation_type
                else:
                    text = self.tr("<b>No annotation data has been detected!</b>")

                self.anotation_detect_label.setText(text)

    def isComplete(self):
        """Conditions to unlock next button"""
        return (
            True
            if (
                self.file_path_edit.text()
                and QFile(self.file_path_edit.text()).exists()
            )
            else False
        )


class SamplePage(QWizardPage):
    """Gather additional information on sequenced individuals and their families

    PED file is facultative, a user can manually fill the data in the dynamic view.

    See Also:
        :meth:`gui.model_view.pedigree.PedView`
    """

    def __init__(self):
        super().__init__()

        self.setTitle(self.tr("Samples"))
        self.setSubTitle(self.tr("Add sample descriptions or skip this step."))
        self.help_text = QLabel(
            self.tr(
                "You can edit the relationships between genomes found in the VCF\n"
                "manually or via a PED/PLINK file (sample pedigree information and "
                "genotype calls).\nBy default the fields are those found in the VCF; "
                "the unknown fields are empty.\nDouble click to edit them."
            )
        )
        self.view = PedView()
        self.import_button = QPushButton(self.tr("Import PED file (facultative)"))
        self.ped_message = QLabel()
        v_layout = QVBoxLayout()
        v_layout.addWidget(self.help_text)
        v_layout.addWidget(self.view)
        v_layout.addWidget(self.import_button)
        v_layout.addWidget(self.ped_message)
        self.setLayout(v_layout)

        self.vcf_samples = list()  # Raw individual_ids from VCF
        # PED data built from VCF ids (other fields than individual_id are default)
        # used as a reference to detect user changes
        self.vcf_default_ped_samples = list()

        self.import_button.clicked.connect(self.on_import_clicked)
        self.view.message.connect(self.on_display_message)

        # Share PedView.pedfile accross Wizard pages
        self.registerField("pedfile", self.view, "pedfile")

    def initializePage(self):
        """Overridden: Prepare the page just before it is shown

        We open variant file (vcf, etc.) to get the current samples that will
        be eventually associated to a PED file later.
        """
        self.view.clear()
        # Open variant file of the project and read its headers
        filename = self.field("filename")
        with create_reader(filename) as reader:
            # Get samples of the vcf project
            self.vcf_samples = reader.get_samples()

            self.vcf_default_ped_samples = [
                # family_id, individual_id, father_id, mother_id, sex, genotype
                ["fam", name, "0", "0", "0", "0", "0"]
                for name in self.vcf_samples
            ]
            # Deepcopy to avoid further modifications of this list of reference
            self.view.samples = copy.deepcopy(self.vcf_default_ped_samples)

    def validatePage(self):
        """Overrided: Called when a user clicks on next button"""
        # Check if PedView contains the same default data
        # print("default", self.vcf_default_ped_samples)
        # print("vs", self.view.samples)
        if set(map(tuple, self.vcf_default_ped_samples)) == set(
            map(tuple, self.view.samples)
        ):
            # Reset samples => will set pedfile field to None
            self.view.samples = list()

        return True

    @Slot()
    def on_import_clicked(self):
        """Slot called when import PED file is clicked

        We open PED file (ped, tfam) to get the their samples that will
        be associated to the current samples of the project.
        """
        # Reload last directory used
        app_settings = QSettings()
        last_directory = app_settings.value("last_directory", QDir.homePath())

        filepath, filetype = QFileDialog.getOpenFileName(
            self,
            self.tr("Open PED file"),
            last_directory,
            self.tr("PED files (*.ped *.tfam)"),
        )

        if not filepath:
            return

        LOGGER.info("vcf samples: %s", self.vcf_samples)

        # Get samples of individual_ids that are already on the VCF file
        self.view.samples = [
            # samples argument is empty dict since its not
            sample
            for sample in PedReader(filepath, dict())
            if sample[1] in self.vcf_samples
        ]

    def on_display_message(self, message):
        """Display messages about data validation from the delegate"""
        self.ped_message.setText(message)


class FieldsPage(QWizardPage):
    """Allow user to skip too import some fields 

    """

    MANDATORY_FIELDS = ["chr", "pos", "ref", "alt", "gt"]

    def __init__(self):
        super().__init__()

        self.setTitle(self.tr("Fields"))
        self.setSubTitle(self.tr("Select fields you want to import"))
        self.help_text = QLabel(self.tr("Check fields you want to import "))
        self.select_button = QPushButton(self.tr("(Un)Select all"))
        self.view = QTableView()
        self.model = QStandardItemModel()
        self.view.setModel(self.model)
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.view.setAlternatingRowColors(True)

        # main layout
        main_layout = QHBoxLayout()
        main_layout.addWidget(self.view)

        self.setLayout(main_layout)

    def initializePage(self):
        """ overload """

        # Load fields
        self.model.clear()
        self.model.setColumnCount(4)
        self.model.setHorizontalHeaderLabels(
            ["name", "category", "description", "type"]
        )

        # Open variant file of the project and read its headers
        filename = self.field("filename")

        with create_reader(filename) as reader:

            for field in reader.get_fields():

                name_item = QStandardItem(field["name"])
                cat_item = QStandardItem(field["category"])
                desc_item = QStandardItem(field["description"])
                type_item = QStandardItem(field["type"])

                name_item.setCheckable(True)
                name_item.setCheckState(Qt.Checked)
                name_item.setData(field)

                line = [name_item, cat_item, desc_item, type_item]

                for item in line:
                    item.setEditable(False)
                    item.setEnabled(not field["name"] in self.MANDATORY_FIELDS)

                self.model.appendRow(line)

        self.view.horizontalHeader().setSectionResizeMode(2, QHeaderView.Stretch)

    def validatePage(self):
        """ override """

        # Loop over fields a create a ignored fields set e.g (("qual","variants"))
        ignored_fields = set()
        for row in range(self.model.rowCount()):
            item = self.model.item(row, 0)
            if item.checkState() == Qt.Unchecked:
                field = item.data()
                ignored_fields.add((field["name"], field["category"]))

        self.wizard().config["ignored_fields"] = ignored_fields

        return True


class ImportThread(QThread):
    """Thread used to create a new project by importing a variant file

    .. note:: This thread creates the database.
    .. seealso:: ImportPage class.
    """

    # Qt signals
    progress_changed = Signal(int, str)
    finished_status = Signal(bool)

    def __init__(self):
        super().__init__()
        self._stop = False
        # File top open
        self.filename = ""
        # Project's filepath
        self.db_filename = ""
        # Facultative PED file
        self.pedfile = None
        # Ignored fields
        self.ignored_fields = set()

        self.project_settings = dict()

    def set_importer_settings(
        self,
        filename,
        db_filename,
        pedfile=None,
        ignored_fields=set(),
        project_settings={},
    ):
        """Init settings of the importer

        :param filename: File to be opened.
        :param db_filename: Filepath of the new project.
        :key pedfile: PED file to be opened.
        :key project_settings: The reference genome and the name of the project.
            Keys have to be at least "reference" and "project_name".
        :type filename: <str>
        :type pedfile: <str>
        :type db_filename: <str>
        :type project_settings: <dict>
        """
        # File top open
        self.filename = filename
        # Project's filepath
        self.db_filename = db_filename
        # Project settings
        self.project_settings = project_settings
        # Ped file
        self.pedfile = pedfile
        # Ignored fields
        self.ignored_fields = ignored_fields

    def run(self):
        """Overrided QThread method

        .. warning:: This method is the only one to be executed in the thread.
            SQLite objects created in a thread can only be used in that same
            thread; so no use of self.conn is allowed elsewhere.
        """
        self._stop = False

        #  start timer
        start = time.perf_counter()

        if os.path.exists(self.db_filename):
            os.remove(self.db_filename)
        self.conn = get_sql_connection(self.db_filename)

        try:
            # Import the file
            for value, message in async_import_file(
                self.conn,
                self.filename,
                pedfile=self.pedfile,
                ignored_fields=self.ignored_fields,
                project=self.project_settings,
            ):
                if self._stop:
                    self.conn.close()
                    break
                # Send progression
                self.progress_changed.emit(value, message)

        except BaseException as e:
            self.progress_changed.emit(0, str(e))
            self._stop = True
            LOGGER.exception(e)
            raise e
        finally:
            # Send status (Send True when there is no error)
            # end timer
            end = time.perf_counter()
            elapsed_time = end - start
            self.progress_changed.emit(100, str("Elapsed time: %.2gs" % (end - start)))
            self.finished_status.emit(not self._stop)

    def stop(self):
        """Stop the import process"""
        self._stop = True


class ImportPage(QWizardPage):
    """Page: Creation of the database"""

    completeChanged = Signal()

    def __init__(self):
        super().__init__()

        self.setTitle(self.tr("Import file"))
        self.setSubTitle(
            self.tr("Please click on Import/Stop to start or stop the process.")
        )
        self.text_buttons = [self.tr("Import"), self.tr("Stop")]

        # Async stuff
        self.thread_finished = False  # True if import process is correctly finished
        self.thread = ImportThread()
        self.progress = QProgressBar()
        self.import_button = QPushButton(self.text_buttons[0])

        v_layout = QHBoxLayout()
        v_layout.addWidget(self.progress)
        v_layout.addWidget(self.import_button)

        m_layout = QVBoxLayout()

        self.log_edit = QPlainTextEdit()
        self.tab_widget = QTabWidget()
        self.tab_widget.addTab(self.log_edit, "log")

        m_layout.addLayout(v_layout)
        m_layout.addWidget(self.tab_widget)
        self.setLayout(m_layout)

        # File to open
        self.log_edit.appendPlainText(self.field("filename"))
        # Database filename; see initializePage()
        self.db_filename = None

        # Ignored field from previous page; see initializePage()
        self.ignored_fields = set()

        self.thread.started.connect(
            lambda: self.log_edit.appendPlainText(self.tr("Started"))
        )

        # Note: self.run is automatically launched when ImportPage is displayed
        # See initializePage()
        self.import_button.clicked.connect(self.run)
        self.thread.progress_changed.connect(self.progress_changed)
        self.thread.finished.connect(self.import_thread_finished)
        self.thread.finished_status.connect(self.import_thread_finished_status)

    def initializePage(self):
        """Overridden: Prepare the page just before it is shown

        Launch import process if currentPage is ImportPage
        """
        # print("Current PED file", self.field("pedfile"))
        self.db_filename = (
            self.field("project_path")
            + QDir.separator()
            + self.field("project_name")
            + ".db"
        )

        self.ignored_fields = self.wizard().config["ignored_fields"]

        self.run()
        self.import_button.setDisabled(False)

    def cleanupPage(self):
        """Called when back button is clicked: stop the import thread"""
        self.thread_stop()

        if self.thread_finished:
            # The import process is not finished corretly
            # Delete file
            os.remove(self.db_filename)

    @Slot()
    def progress_changed(self, value, message):
        """Update the progress bar
        Slot called when progress_changed is emitted by the thread
        """
        self.progress.setValue(value)
        if message:
            self.log_edit.appendPlainText(message)

    @Slot()
    def import_thread_finished(self):
        """Force the activation of the finish button after a successful import"""
        try:
            self.completeChanged.emit()
        except RuntimeError:
            # When closing the wizard, the thread is stopped via cleanupPage()
            # and finished signal is emitted after the deletion of the wizard.
            pass

    @Slot()
    def import_thread_finished_status(self, status):
        """Set the finished status of import thread

        .. note:: Called at the end of run()
        """
        self.thread_finished = status
        if status:
            # Block further import
            self.import_button.setDisabled(True)
            self.log_edit.appendPlainText(self.tr("Done"))
        else:
            # Display import on the button
            self.import_button.setText(self.text_buttons[0])
            self.log_edit.appendPlainText(self.tr("Stopped!"))

    @Slot()
    def run(self):
        """Called when import button is clicked
        Launch the import in a separate thread
        """
        if self.thread.isRunning():
            self.thread_stop()
        else:
            self.thread.set_importer_settings(
                # File to open
                filename=self.field("filename"),
                # Project's filepath
                db_filename=self.db_filename,
                # PED file to use with samples
                pedfile=self.field("pedfile"),
                # Ignored fields
                ignored_fields=self.ignored_fields,
                # Project's settings
                project_settings={
                    # Reference genome
                    "reference": self.field("reference"),
                    # Project's name
                    "project_name": self.field("project_name"),
                },
            )

            self.log_edit.appendPlainText(self.tr("Import ") + self.thread.filename)

            show_ignored_fields = ",".join([i[0] for i in self.ignored_fields])
            self.log_edit.appendPlainText("Ignored fields: " + show_ignored_fields)
            # display stop on the button
            self.import_button.setText(self.text_buttons[1])
            self.thread.start()

    def thread_stop(self):
        """Stop the thread and refresh the UI"""
        LOGGER.debug("ImportPage:run: stop thread")
        self.import_button.setDisabled(True)
        self.thread.stop()
        self.progress.setValue(0)
        self.import_button.setDisabled(False)
        self.import_button.setText(self.text_buttons[0])

    def isComplete(self):
        """Conditions to unlock finish button"""
        return self.thread_finished


class ProjectWizard(QWizard):
    """Main window of the project wizard

    3 pages are instantiated here:
        - ProjectPage: Creation of a new project
        - FilePage: Open the file containing variant data
        - ImportPage: Creation of the database
    """

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Cutevariant - Project creation wizard"))
        self.setWindowIcon(QIcon(cm.DIR_ICONS + "app.png"))
        self.setWizardStyle(QWizard.ClassicStyle)
        self.addPage(ProjectPage())
        self.addPage(FilePage())
        self.addPage(SamplePage())
        self.addPage(FieldsPage())
        self.addPage(ImportPage())

        # Stored all data filled by the wizard
        # Better than using cumberstome setField...
        self.config = {}


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)

    w = ProjectWizard()
    w.show()

    app.exec_()
